import os.path

import util.index_database.general
import util.index_database.array_based
import util.index_database.txt_file_based

import util.logging
logger = util.logging.logger



class Database(util.index_database.general.Database):
    
    def __init__(self, array_file, value_file, value_reliable_decimal_places=18, tolerance_options=None):
        ## call super constructor
        super().__init__(value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
        ## create array and file db
        self.array_db = util.index_database.array_based.Database(array_file, value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
        value_dir, value_filename = os.path.split(value_file)
        self.txt_file_db = util.index_database.txt_file_based.Database(value_dir, value_filename, value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
    
    
    def __str__(self):
        return 'Index array and file system database'
    
    
    ## setter and getter for files
    
    @property
    def array_file(self):
        return self.array_db.array_file
    
    @array_file.setter
    def array_file(self, array_file): 
        self.array_db.array_file = array_file
    
    
    @property
    def value_file(self):
        return os.path.join(self.txt_file_db.value_dir, self.txt_file_db.value_filenames[0])
    
    @value_file.setter
    def value_file(self, value_file): 
        value_dir, value_filename = os.path.split(value_file)
        if value_dir != self.txt_file_db.value_dir:
            self.txt_file_db.value_dir = value_dir
        if value_filename != self.txt_file_db.value_filenames[0]:
            self.txt_file_db.value_filenames[0] = value_filename

    
    ## access
    
    def get_value(self, index):
        return self.array_db.get_value(index)

    def has_value(self, index):
        return self.array_db.has_value(index)

    def set_value(self, index, value, overwrite=False):
        logger.debug('{}: Setting value at index {} to {} with overwrite {}.'.format(self, index, value, overwrite))
        self.txt_file_db.set_value(index, value, overwrite=overwrite)
        self.array_db.set_value(index, value, overwrite=overwrite)

    def add_value(self, value):
        logger.debug('{}: Adding value {}'.format(self, value))
        with self.array_db.locked_file.lock_object(exclusive=True):
            index = self.array_db.add_value(value)
            try:
                self.txt_file_db.set_value(index, value, overwrite=False)
            except util.index_database.general.DatabaseIndexError as e:
                self.array_db.remove_index(index)
                raise e
        return index
        
    
    def used_indices(self):
        return self.array_db.used_indices()
    
    def remove_index(self, index, force=False):
        logger.debug('{}: Removing index {}.'.format(self, index))
        self.txt_file_db.remove_index(index, force=force)
        self.array_db.remove_index(index)
    
    def closest_indices(self, value):
        return self.array_db.closest_indices(value)

    def closest_index(self, value):
        return self.array_db.closest_index(value)

    def index(self, value):
        return self.array_db.index(value)


    ## merge
    
    def merge_txt_file_db_to_array_db(self):
        logger.debug('{}: Merging file db to array db.'.format(self))
        for index in self.txt_file_db.used_indices():
            txt_file_db_value = self.txt_file_db.get_value(index)
            if not self.array_db.has_value(index) or not self.are_values_equal(self.array_db.get_value(index), txt_file_db_value):
                self.array_db.set_value(index, txt_file_db_value, overwrite=False)
        



import abc
import math
import os
import subprocess
import time

import numpy as np

import util.io.fs
import util.options

import util.logging
logger = util.logging.logger


class NodeInfos():
    
    def __init__(self, node_infos):
        self.node_infos = node_infos
    
    def kinds(self):
        return tuple(self.node_infos.keys())

    def nodes(self, kind):
        return self.node_infos[kind]['nodes']

    def cpus(self, kind):
        return self.node_infos[kind]['cpus']

    def speed(self, kind):
        return self.node_infos[kind]['speed']

    def memory(self, kind):
        return self.node_infos[kind]['memory']

    def leave_free(self, kind):
        node_info_kind = self.node_infos[kind]
        try:
            return node_info_kind['leave_free']
        except KeyError:
            return 0

    def max_walltime(self, kind):
        node_info_kind = self.node_infos[kind]
        try:
            return node_info_kind['max_walltime']
        except KeyError:
            return float('inf')



class NodesState():
    
    def __init__(self, nodes_state):
        self.nodes_state = nodes_state
    
    def nodes_state_for_kind(self, kind):
        nodes_state_values_for_kind = self.nodes_state_values_for_kind(kind)
        return NodesState({kind: nodes_state_values_for_kind})
    
    def nodes_state_values_for_kind(self, kind):
        try:
            nodes_state_values_for_kind = self.nodes_state[kind]
        except KeyError as e:
            logger.warning('Node kind {} not found in nodes state {}.'.format(kind, self.nodes_state))
            nodes_state_values_for_kind = [np.array([]), np.array([])]
        return nodes_state_values_for_kind
    
    def free_cpus(self, kind, required_memory=0):
        if required_memory == 0:
            free_cpus = self.nodes_state_values_for_kind(kind)[0]
        else:
            free_memory = self.free_memory(kind)
            free_cpus = self.free_cpus(kind, required_memory=0)
            free_cpus = free_cpus[free_memory >= required_memory]
        return free_cpus

    def free_memory(self, kind):
        return self.nodes_state_values_for_kind(kind)[1]



class NodeSetup:

    def __init__(self, memory=None, node_kind=None, nodes=None, cpus=None, nodes_max=float('inf'), nodes_leave_free=0, total_cpus_min=1, total_cpus_max=float('inf'), check_for_better=False, walltime=None):
        
        ## set batch system
        from util.batch.universal.system import BATCH_SYSTEM
        self.batch_system = BATCH_SYSTEM
        
        ## check input
        assert nodes is None or nodes >= 1
        assert cpus is None or cpus >= 1
        assert total_cpus_max is None or total_cpus_min is None or total_cpus_max >= total_cpus_min
        assert nodes_max is None or nodes is None or nodes_max >= nodes
        assert total_cpus_min is None or nodes is None or cpus is None or total_cpus_min <= nodes * cpus
        assert total_cpus_max is None or nodes is None or cpus is None or total_cpus_max >= nodes * cpus
        assert total_cpus_max is None or nodes is None or total_cpus_max >= nodes
        assert total_cpus_max is None or cpus is None or total_cpus_max >= cpus
        
        if node_kind is not None and cpus is not None:
            max_cpus = self.batch_system.node_infos.cpus(node_kind)
            if cpus > max_cpus:
                raise ValueError('For node kind {} are maximal {} cpus per node available but {} are reqeusted'.format(node_kind, max_cpus, cpus))      
        
        if node_kind is not None and nodes is not None:
            max_nodes = self.batch_system.node_infos.nodes(node_kind)
            if nodes > max_nodes:
                raise ValueError('For node kind {} are maximal {} nodes available but {} are reqeusted'.format(node_kind, max_nodes, nodes))

        ## prepare input
        if node_kind is not None and not isinstance(node_kind, str) and len(node_kind) == 1:
            node_kind = node_kind[0]
        if nodes_max == 1 and nodes is None:
            nodes = 1
        if nodes is not None and total_cpus_max == nodes:
            cpus = 1

        ## save setup
        setup = {'memory': memory, 'node_kind': node_kind, 'nodes': nodes, 'cpus': cpus, 'nodes_max': nodes_max, 'nodes_leave_free': nodes_leave_free, 'total_cpus_min': total_cpus_min, 'total_cpus_max': total_cpus_max, 'check_for_better': check_for_better, 'walltime': walltime}
        self.setup = setup
        

    def __getitem__(self, key):
        return self.setup[key]

    def __setitem__(self, key, value):
        self.setup[key] = value

    def __str__(self):
        dict_str = str(self.setup).replace(": inf", ": float('inf')")
        return '{}(**{})'.format(self.__class__.__name__, dict_str)
    
    def __repr__(self):
        dict_str = str(self.setup).replace(": inf", ": float('inf')")
        return '{}.{}(**{})'.format(self.__class__.__module__, self.__class__.__name__, dict_str)
        

    def __copy__(self):
        copy = type(self)(**self.setup)
        return copy

    def copy(self):
        return self.__copy__()


    def configuration_is_complete(self):
        return self['memory'] is not None and self['node_kind'] is not None and isinstance(self['node_kind'], str) and self['nodes'] is not None and self['cpus'] is not None


    def complete_configuration(self):
        if not self.configuration_is_complete():
            logger.debug('Node setup incomplete. Try to complete it.')
            if self['memory'] is None:
                raise ValueError('Memory has to be set.')
            try:
                (node_kind, nodes, cpus) = self.batch_system.wait_for_needed_resources(self['memory'], node_kind=self['node_kind'], nodes=self['nodes'], cpus=self['cpus'], nodes_max=self['nodes_max'], nodes_leave_free=self['nodes_leave_free'], total_cpus_min=self['total_cpus_min'], total_cpus_max=self['total_cpus_max'])
            except NotImplementedError:
                logger.error('Batch system does not support completion of node setup.')
                raise NodeSetupIncompleteError(self)
            self['node_kind'] = node_kind
            self['nodes'] = nodes
            self['cpus'] = cpus


    def configuration_value(self, key, test=None):
        assert test is None or callable(test)

        value = self.setup[key]
        if value is None or (test is not None and not test(value)):
            self.complete_configuration()
            value = self.setup[key]

        assert value is not None
        return value


    def update_with_best_configuration(self, check_for_better=True, not_free_speed_factor=0.7):
        if check_for_better:
            self['check_for_better'] = False
            setup_triple = (self.node_kind, self.nodes, self.cpus)
            logger.debug('Try to find better node setup configuration than {}.'.format(setup_triple))
            speed = self.batch_system.speed(*setup_triple)
            best_setup_triple = self.batch_system.best_cpu_configurations(self.memory, nodes_max=self['nodes_max'], total_cpus_max=self['total_cpus_max'], walltime=self.walltime)
            best_speed = self.batch_system.speed(*best_setup_triple)
            if best_speed > speed:
                logger.debug('Using better node setup configuration {}.'.format(best_setup_triple))
                self['node_kind'], self['nodes'], self['cpus'] = best_setup_triple
            elif not self.batch_system.is_free(self.memory, self.node_kind, self.nodes, self.cpus):
                logger.debug('Node setup configuration {} is not free.'.format(setup_triple))
                if best_speed >= speed * not_free_speed_factor:
                    logger.debug('Using node setup configuration {}.'.format(best_setup_triple))
                    self['node_kind'], self['nodes'], self['cpus'] = best_setup_triple
                else:
                    logger.debug('Not using best node setup configuration {} since it is to slow.'.format(best_setup_triple))


    @property
    def memory(self):
        return self.setup['memory']
    
    @memory.setter
    def memory(self, memory):
        self.setup['memory'] = memory
    

    @property
    def node_kind(self):
        self.update_with_best_configuration(self['check_for_better'])
        return self.configuration_value('node_kind', test=lambda v: isinstance(v, str))

    @property
    def nodes(self):
        self.update_with_best_configuration(self['check_for_better'])
        return self.configuration_value('nodes')

    @property
    def cpus(self):
        self.update_with_best_configuration(self['check_for_better'])
        return self.configuration_value('cpus')


    @property
    def walltime(self):
        return self.setup['walltime']
    
    @walltime.setter
    def walltime(self, walltime):
        self.setup['walltime'] = walltime


    @property
    def total_cpus_min(self):
        return self.setup['total_cpus_min']
    
    @total_cpus_min.setter
    def total_cpus_min(self, total_cpus_min):
        self.setup['total_cpus_min'] = total_cpus_min


    @property
    def nodes_max(self):
        return self.setup['nodes_max']
    
    @nodes_max.setter
    def nodes_max(self, nodes_max):
        self.setup['nodes_max'] = nodes_max



class NodeSetupIncompleteError(Exception):

    def __init__(self, nodes_setup):
        error_message = 'The node setup is incomplete: node_kind={}, nodes={} and cpus={}.'.format(nodes_setup.node_kind, nodes_setup.nodes, nodes_setup.cpus)
        super().__init__(error_message)




class BatchSystem():

    def __init__(self, commands, queues, max_walltime={}, module_renaming={}, node_infos={}):
        self.commands = commands
        logger.debug('{} initiating with commands {}, queues {}, max_walltime {} and module_renaming {}.'.format(self, commands, queues, max_walltime, module_renaming))
        self.queues = queues
        self.max_walltime = max_walltime
        self.module_renaming = module_renaming
        
        if not isinstance(node_infos, NodeInfos):
            node_infos = NodeInfos(node_infos)
        self.node_infos = node_infos


    @property
    def mpi_command(self):
        return self.commands['mpirun']

    @property
    def time_command(self):
        return self.commands['time']

    @property
    def submit_command(self):
        return self.commands['sub']

    @property
    def status_command(self):
        return self.commands['stat']

    @property
    def nodes_command(self):
        return self.commands['nodes']


    def __str__(self):
        return 'General batch system'

    ## check methods

    def check_queue(self, queue):
        if queue is not None and queue not in self.queues:
            raise ValueError('Unknown queue {}.'.format(queue))
        return queue


    def check_walltime(self, queue, walltime_hours):
        ## get max walltime
        try:
            max_walltime_for_queue = self.max_walltime[queue]
        except KeyError:
            max_walltime_for_queue = float('inf')
        ## check walltime
        if walltime_hours is not None:
            if walltime_hours <= max_walltime_for_queue:
                walltime_hours = math.ceil(walltime_hours)
            else:
                raise ValueError('Max walltime {} is greater than max walltime for queue {}.'.format(walltime_hours, max_walltime_for_queue))
        else:
            if max_walltime_for_queue < float('inf'):
                walltime_hours = max_walltime_for_queue
        ## return
        assert (walltime_hours is None and max_walltime_for_queue == float('inf')) or walltime_hours <= max_walltime_for_queue
        return walltime_hours


    def check_modules(self, modules):
        if len(modules) > 0:
            modules = list(modules)

            ## add required modules
            for module_requested, modules_required in [('petsc', ('intelmpi', 'intel')), ('intelmpi', ('intel',)), ('hdf5', ('intel',))]:
                if module_requested in modules:
                    for module_required in modules_required:
                        if module_required not in modules:
                            logger.warning('Module "{}" needs module "{}" to be loaded first.'.format(module_requested, module_required))
                            modules = [module_required] + modules

            ## check positions
            first_index = 0
            for module in  ('intel', 'intelmpi'):
                if module in modules and modules[first_index] != module:
                    logger.warning('Module "{}" has to be at position {}.'.format(module, first_index))
                    modules.remove(module)
                    modules = [module] + modules
                first_index += 1

            ## rename modules
            for i in range(len(modules)):
                module = modules[i]
                try:
                    module_new = self.module_renaming[module]
                except KeyError:
                    module_new = module
                modules[i] = module_new
        return modules


    ## other methods

    def start_job(self, job_file):
        logger.debug('Starting job with option file {}.'.format(job_file))

        if not os.path.exists(job_file):
            raise FileNotFoundError(job_file)

        submit_output = subprocess.check_output((self.submit_command, job_file)).decode("utf-8")
        submit_output = submit_output.strip()
        job_id = self._get_job_id_from_submit_output(submit_output)

        logger.debug('Started job has ID {}.'.format(job_id))

        return job_id


    def is_mpi_used(self, modules):
        for module in modules:
            if 'intelmpi' in module:
                return True
        return False


    def add_mpi_to_command(self, command, cpus, use_mpi=True):
        if use_mpi:
            command = self.mpi_command.format(command=command, cpus=cpus)
        return command
    
    
    ## best node setups
    
    def speed(self, node_kind, nodes, cpus):
        return self.node_infos.speed(node_kind) * nodes * cpus
    
    
    def is_free(self, memory, node_kind, nodes, cpus):
        ## get nodes with required memory
        nodes_state = self._nodes_state()
        free_cpus = nodes_state.free_cpus(node_kind, required_memory=memory)
        
        ## calculate useable nodes
        free_nodes = free_cpus[free_cpus >= cpus].size
        free_nodes = free_nodes - self.node_infos.leave_free(node_kind)
        
        return free_nodes >= nodes
    
    
    @staticmethod
    def _best_cpu_configurations_for_state(nodes_state, node_kind, memory_required, nodes=None, cpus=None, nodes_max=float('inf'), nodes_leave_free=0, total_cpus_max=float('inf')):
        logger.debug('Getting best cpu configuration for node state {} with memory {}, nodes {}, cpus {}, nodes max {} and nodes left free {}.'.format(nodes_state, memory_required, nodes, cpus, nodes_max, nodes_leave_free))
    
        ## check input
        if nodes_max <= 0:
            raise ValueError('nodes_max {} has to be greater 0.'.format(nodes_max))
        if total_cpus_max <= 0:
            raise ValueError('total_cpus_max {} has to be greater 0.'.format(total_cpus_max))
        if nodes_leave_free < 0:
            raise ValueError('nodes_leave_free {} has to be greater or equal to 0.'.format(nodes_leave_free))
        if nodes is not None:
            if nodes <= 0:
                raise ValueError('nodes {} has to be greater 0.'.format(nodes))
            if nodes > nodes_max:
                raise ValueError('nodes_max {} has to be greater or equal to nodes {}.'.format(nodes_max, nodes))
        if cpus is not None:
            if cpus <= 0:
                raise ValueError('cpus {} has to be greater 0.'.format(cpus))
        if nodes is not None and cpus is not None:
            if nodes * cpus > total_cpus_max:
                raise ValueError('total_cpus_max {} has to be greater or equal to nodes {} multiplied with cpus {}.'.format(total_cpus_max, nodes, cpus))
    
        ## get only nodes with required memory
        free_cpus = nodes_state.free_cpus(node_kind, required_memory=memory_required)
    
        ## calculate best configuration
        best_nodes = 0
        best_cpus = 0
    
        if len(free_cpus) > 0:
            ## chose numbers of cpus to check
            if cpus is not None:
                cpus_to_check = (cpus,)
            else:
                cpus_to_check = range(max(free_cpus), 0, -1)
    
            ## get number of nodes for each number of cpus
            for cpus_to_check_i in cpus_to_check:
                ## calculate useable nodes (respect max nodes and left free nodes)
                free_nodes = free_cpus[free_cpus >= cpus_to_check_i].size
                free_nodes = free_nodes - nodes_leave_free
                free_nodes = min(free_nodes, nodes_max)
    
                ## respect fix number of nodes if passed
                if nodes is not None:
                    if free_nodes >= nodes:
                        free_nodes = nodes
                    else:
                        free_nodes = 0
    
                ## respect total max cpus
                while free_nodes * cpus_to_check_i > total_cpus_max:
                    if free_nodes > 1:
                        free_nodes -= 1
                    else:
                        cpus_to_check_i = total_cpus_max
    
                ## check if best configuration
                if free_nodes * cpus_to_check_i > best_nodes * best_cpus:
                    best_nodes = free_nodes
                    best_cpus = cpus_to_check_i
    
        logger.debug('Best CPU configuration is for this kind: {}'.format((best_nodes, best_cpus)))
    
        assert best_nodes <= nodes_max
        assert best_nodes * best_cpus <= total_cpus_max
        assert nodes is None or best_nodes == nodes or best_nodes == 0
        assert cpus is None or best_cpus == cpus or best_cpus == 0
        return (best_nodes, best_cpus)

    
    def best_cpu_configurations(self, memory_required, node_kind=None, nodes=None, cpus=None, nodes_max=float('inf'), nodes_leave_free=0, total_cpus_max=float('inf'), walltime=None):

        logger.debug('Calculating best CPU configurations for {}GB memory with node kinds {}, nodes {}, cpus {}, nodes_max {}, nodes_leave_free {}, total_cpus_max {} and walltime {}'.format(memory_required, node_kind, nodes, cpus, nodes_max, nodes_leave_free, total_cpus_max, walltime))
    
        ## chose node kinds if not passed
        if node_kind is None:
            if walltime is None:
                walltime = 0
            node_kind = []
            for node_kind_i in self.node_infos.kinds():
                if self.node_infos.nodes(node_kind_i) > self.node_infos.leave_free(node_kind_i) and self.node_infos.max_walltime(node_kind_i) >= walltime:
                    node_kind.append(node_kind_i)
        elif isinstance(node_kind, str):
            node_kind = (node_kind,)
        nodes_state = self._nodes_state()
    
        ## init
        best_kind = node_kind[0]
        best_nodes = 0
        best_cpus = 0
        best_cpu_power = 0
    
        ## calculate best CPU configuration
        for node_kind_i in node_kind:
            nodes_cpu_power_i = self.node_infos.speed(node_kind_i)
            nodes_max_i = self.node_infos.nodes(node_kind_i)
            nodes_max_i = min(nodes_max, nodes_max_i)
            nodes_leave_free_i = self.node_infos.leave_free(node_kind_i)
            nodes_leave_free_i = max(nodes_leave_free, nodes_leave_free_i)
            
            (best_nodes_i, best_cpus_i) = self._best_cpu_configurations_for_state(nodes_state, node_kind_i, memory_required, nodes=nodes, cpus=cpus, nodes_max=nodes_max_i, nodes_leave_free=nodes_leave_free_i, total_cpus_max=total_cpus_max)

            logger.debug('Best CPU configurations for {}GB memory with node kind {}, nodes {}, cpus {}, nodes_max {}, nodes_leave_free {} and total_cpus_max {} is {}.'.format(memory_required, node_kind_i, nodes, cpus, nodes_max, nodes_leave_free, total_cpus_max, (best_nodes_i, best_cpus_i)))
    
            if nodes_cpu_power_i * best_cpus_i * best_nodes_i > best_cpu_power * best_cpus * best_nodes:
                best_kind = node_kind_i
                best_nodes = best_nodes_i
                best_cpus = best_cpus_i
                best_cpu_power = nodes_cpu_power_i
        
        ## return
        best_configuration = (best_kind, best_nodes, best_cpus)
    
        logger.debug('Best CPU configuration is: {}.'.format(best_configuration))
    
        assert best_kind in node_kind
        assert best_nodes <= nodes_max
        assert best_nodes * best_cpus <= total_cpus_max
        return best_configuration

    
    def wait_for_needed_resources(self, memory_required, node_kind=None, nodes=None, cpus=None, nodes_max=float('inf'), nodes_leave_free=0, total_cpus_min=1, total_cpus_max=float('inf')):
        logger.debug('Waiting for at least {} CPUs with {}GB memory, with node_kind {}, nodes {}, cpus {}, nodes_max {}, nodes_leave_free {}, total_cpus_min {} and total_cpus_max {}.'.format(total_cpus_min, memory_required, node_kind, nodes, cpus, nodes_max, nodes_leave_free, total_cpus_min, total_cpus_max))
    
        ## check input
        if total_cpus_min > total_cpus_max:
            raise ValueError('total_cpus_max has to be greater or equal to total_cpus_min, but {} < {}.'.format(total_cpus_max, total_cpus_min))
    
        ## calculate
        best_nodes = 0
        best_cpus = 0
        resources_free = False
        while not resources_free:
            (best_cpu_kind, best_nodes, best_cpus) = self.best_cpu_configurations(memory_required, node_kind=node_kind, nodes=nodes, cpus=cpus, nodes_max=nodes_max, nodes_leave_free=nodes_leave_free, total_cpus_max=total_cpus_max)
            cpus_avail = best_nodes * best_cpus
            resources_free = (cpus_avail >= total_cpus_min)
            if not resources_free:
                logger.debug('No enough resources free. {} CPUs available, but {} CPUs needed. Waiting ...'.format(cpus_avail, total_cpus_min))
                time.sleep(60)
    
        return (best_cpu_kind, best_nodes, best_cpus)


    ## abstract methods

    @abc.abstractmethod
    def _get_job_id_from_submit_output(self, submit_output):
        raise NotImplementedError()

    @abc.abstractmethod
    def _nodes_state(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def is_job_running(self, job_id):
        raise NotImplementedError()



class Job():

    def __init__(self, batch_system, output_dir, force_load=False, max_job_name_len=80, exceeded_walltime_error_message=None):
        ## batch system
        self.batch_system = batch_system
        self.max_job_name_len = max_job_name_len
        self.exceeded_walltime_error_message = exceeded_walltime_error_message

        ## check input
        if output_dir is None:
            raise ValueError('The output dir is not allowed to be None.')
        output_dir_expanded = os.path.expandvars(output_dir)
        
        ## get option file
        try:
            option_file_expanded = os.path.join(output_dir_expanded, 'job_options.hdf5')
        except Exception as e:
            raise ValueError('The output dir {} is not allowed.'.format(output_dir)) from e

        ## load option file if existing or forced
        if force_load or os.path.exists(option_file_expanded):
            self.__options = util.options.OptionsFile(option_file_expanded, mode='r+', replace_environment_vars_at_get=True)
            logger.debug('Job {} loaded.'.format(option_file_expanded))

        ## make new job options file otherwise
        else:
            os.makedirs(output_dir_expanded, exist_ok=True)

            self.__options = util.options.OptionsFile(option_file_expanded, mode='w-', replace_environment_vars_at_get=True)

            self.options['/job/output_file'] = os.path.join(output_dir, 'job_output.txt')
            self.options['/job/option_file'] = os.path.join(output_dir, 'job_options.txt')
            self.options['/job/id_file'] = os.path.join(output_dir, 'job_id.txt')
            self.options['/job/unfinished_file'] = os.path.join(output_dir, 'unfinished.txt')
            self.options['/job/finished_file'] = os.path.join(output_dir, 'finished.txt')

            logger.debug('Job {} initialized.'.format(option_file_expanded))


    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


    def __str__(self):
        output_dir = self.output_dir
        try:
            job_id = self.id
        except KeyError:
            job_id = None
        if job_id is not None:
            job_str = 'job {} with output path {}'.format(job_id, output_dir)
        else:
            job_str = 'not started job with output path {}'.format(job_id, output_dir)
        return job_str


    @property
    def options(self):
        return self.__options


    ## option properties

    def option_value(self, name, not_exist_okay=True, replace_environment_vars=True):
        replace_environment_vars_old = self.options.replace_environment_vars_at_get
        self.options.replace_environment_vars_at_get = replace_environment_vars
        try:
            if not_exist_okay:
                try:
                    return self.options[name]
                except KeyError:
                    return None
            else:
                return self.options[name]
        finally:
            self.options.replace_environment_vars_at_get = replace_environment_vars_old


    @property
    def id(self):
        try:
            return self.options['/job/id']
        except KeyError:
            raise KeyError('Job with option file ' + self.options.filename + ' is not started!')

    @property
    def output_dir(self):
        return os.path.dirname(self.option_value('/job/output_file', False))

    @property
    def output_dir_not_expanded(self):
        return os.path.dirname(self.option_value('/job/output_file', False, replace_environment_vars=False))

    @property
    def option_file(self):
        return self.option_value('/job/option_file', False)

    @property
    def unfinished_file(self):
        return self.option_value('/job/unfinished_file', False)

    @property
    def finished_file(self):
        return self.option_value('/job/finished_file', False)

    @property
    def id_file(self):
        return self.option_value('/job/id_file', True)

    @property
    def output_file(self):
        return self.option_value('/job/output_file', True)

    @property
    def output(self):
        output_file = self.output_file
        if output_file is not None and os.path.exists(output_file):
            with open(output_file, 'r') as file:
                output = file.read()
        else:
            output = None
        return output

    @property
    def exit_code(self):
        ## check if finished file exists
        if not os.path.exists(self.finished_file):
            ValueError('Finished file {} does not exist. The job is not finished'.format(self.finished_file))
        ## read exit code
        with open(self.finished_file, mode='r') as finished_file:
            exit_code = finished_file.read()
        ## check exit code
        if len(exit_code) > 0:
            try:
                exit_code = int(exit_code)
                return exit_code
            except ValueError:
                raise ValueError('Finished file {} does not contain an exit code but rather {}.'.format(self.finished_file, exit_code))
        else:
            raise ValueError('Finished file {} is empty.'.format(self.finished_file))

    @property
    def cpu_kind(self):
        return self.option_value('/job/cpu_kind', True)

    @property
    def nodes(self):
        return self.option_value('/job/nodes', True)

    @property
    def cpus(self):
        return self.option_value('/job/cpus', True)

    @property
    def queue(self):
        return self.option_value('/job/queue', True)

    @property
    def walltime_hours(self):
        return self.option_value('/job/walltime_hours', True)


    ## init methods

    def init_job_file(self, job_name, nodes_setup, queue=None, cpu_kind=None):
        ## check qeue and walltime
        queue = self.batch_system.check_queue(queue)
        walltime_hours = nodes_setup.walltime
        walltime_hours = self.batch_system.check_walltime(queue, walltime_hours)

        ## set job options
        self.options['/job/memory_gb'] = nodes_setup.memory
        self.options['/job/nodes'] = nodes_setup.nodes
        self.options['/job/cpus'] = nodes_setup.cpus
        self.options['/job/queue'] = queue
        self.options['/job/name'] = job_name[:self.max_job_name_len]
        if cpu_kind is not None:
            self.options['/job/cpu_kind'] = cpu_kind
        if walltime_hours is not None:
            self.options['/job/walltime_hours'] = walltime_hours


    @abc.abstractmethod
    def _make_job_file_header(self, use_mpi):
        raise NotImplementedError()

    @abc.abstractmethod
    def _make_job_file_modules(self, modules):
        raise NotImplementedError()

    def _make_job_file_command(self, run_command, pre_run_command=None, add_timing=True):
        if add_timing:
            run_command = self.batch_system.time_command.format(command=run_command)
        content = []
        content.append('touch {}'.format(self.options['/job/unfinished_file']))
        content.append('echo "Job started."')
        content.append('')
        if pre_run_command is not None:
            content.append(pre_run_command)
        content.append(run_command)
        content.append('')
        content.append('EXIT_CODE=$?')
        content.append('echo "Job finished with exit code $EXIT_CODE."')
        content.append('rm {}'.format(self.options['/job/unfinished_file']))
        content.append('echo $EXIT_CODE > {}'.format(self.options['/job/finished_file']))
        content.append('exit')
        content.append('')
        return os.linesep.join(content)


    def write_job_file(self, run_command, pre_run_command=None, modules=()):
        modules = self.batch_system.check_modules(modules)
        use_mpi = self.batch_system.is_mpi_used(modules)
        cpus = self.options['/job/nodes'] * self.options['/job/cpus']
        run_command = self.batch_system.add_mpi_to_command(run_command, cpus, use_mpi=use_mpi)
        with open(self.option_file, mode='w') as file:
            file.write(self._make_job_file_header(use_mpi=use_mpi))
            file.write(self._make_job_file_modules(modules))
            file.write(self._make_job_file_command(run_command, pre_run_command=pre_run_command))


    ## other methods

    def start(self):
        job_id = self.batch_system.start_job(self.options['/job/option_file'])
        self.options['/job/id'] = job_id

        id_file = self.id_file
        if id_file is not None:
            with open(self.options['/job/id_file'], 'w') as id_file:
                id_file.write(job_id)


    def is_started(self):
        try:
            self.options['/job/id']
            return True
        except KeyError:
            return False


    def is_finished(self, check_exit_code=True):
        ## if finished file exists, check exit code and output file
        if os.path.exists(self.finished_file):
            if check_exit_code:
                exit_code = self.exit_code
                if exit_code != 0:
                    raise JobExitCodeError(self.id, self.output_dir, exit_code, self.output)
            return self.output_file is None or os.path.exists(self.output_file)
        
        ## if finished file odes not exist, check if running
        elif self.is_started() and not self.batch_system.is_job_running(self.id) and not os.path.exists(self.finished_file):
            time.sleep(60)
            if os.path.exists(self.finished_file):
                return self.is_finished(check_exit_code=check_exit_code)
            else:
                output = self.output
                if self.exceeded_walltime_error_message is not None and self.exceeded_walltime_error_message in output:
                    raise JobExceededWalltimeError(self.id, self.output_dir, self.walltime_hours, output)
                else:
                    raise JobError(self.id, self.output_dir, 'The job is not finished but it is not running! The finished file {} is missing'.format(self.finished_file), output)
        
        ## if not not started or running, return false
        else:
            return False


    def is_running(self):
        return self.is_started() and not self.is_finished(check_exit_code=False)


    def wait_until_finished(self, check_exit_code=True, pause_seconds=None, pause_seconds_min=5, pause_seconds_max=60, pause_seconds_increment_cycle=50):
        adaptive = pause_seconds is None
        if adaptive:
            logger.debug('Waiting for job {} to finish with adaptive sleep period with min {} and max {} seconds and increment cycle {}.'.format(self.id, pause_seconds_min, pause_seconds_max, pause_seconds_increment_cycle))
            pause_seconds = pause_seconds_min
        else:
            logger.debug('Waiting for job {} to finish with {}s sleep period.'.format(self.id, pause_seconds))

        cycle = 0
        while not self.is_finished(check_exit_code=check_exit_code):
            time.sleep(pause_seconds)

            if adaptive:
                cycle += 1
                if cycle == pause_seconds_increment_cycle:
                    pause_seconds += 1
                    cycle = 0

        logger.debug('Job {} finished with exit code {}.'.format(self.id, self.exit_code))


    def make_read_only_input(self, read_only=True):
        if read_only:
            self.options.make_read_only()
            util.io.fs.make_read_only(self.option_file)
            util.io.fs.make_read_only(self.id_file)

    def make_read_only_output(self, read_only=True):
        if read_only:
            if self.output_file is not None:
                util.io.fs.make_read_only(self.output_file)
            util.io.fs.make_read_only(self.finished_file)

    def make_read_only(self, read_only=True):
        self.make_read_only_input(read_only=read_only)
        self.make_read_only_output(read_only=read_only)


    def close(self):
        try:
            options = self.__options
        except AttributeError:
            options = None

        if options is not None:
            options.close()



class JobError(Exception):
    def __init__(self, job_id, output_path, error_message, ouput=None):
        error_message = 'An error accured in job {} stored at {}: {}'.format(job_id, output_path, error_message)
        if ouput is not None:
            error_message = error_message + '\nThe job output was:\n{}'.format(ouput)
        super().__init__(error_message)

class JobExitCodeError(JobError):
    def __init__(self, job_id, output_path, exit_code, ouput=None):
        error_message = 'The command of the job exited with code {}.'.format(exit_code)
        super().__init__(job_id, output_path, error_message, ouput=ouput)


class JobExceededWalltimeError(JobError):
    def __init__(self, job_id, output_path, walltime, ouput=None):
        error_message = 'The job {} exceeded walltime {}.'.format(job_id, walltime)
        super().__init__(job_id, output_path, error_message, ouput=ouput)



import numpy as np
import scipy.sparse
import warnings

import util.logging
logger = util.logging.logger


def squared(A):
    assert A.ndim == 2 and A.shape[0] == A.shape[1]


def sorted_squared_csr(A):
    squared(A)

    if not scipy.sparse.isspmatrix_csr(A):
        warnings.warn('CSR matrix format is required. Converting to CSR matrix.', scipy.sparse.SparseEfficiencyWarning)
        A = scipy.sparse.csr_matrix(A)

    A.sort_indices()
    A.eliminate_zeros()
    return A


def sorted_squared_csc(A):
    squared(A)

    if not scipy.sparse.isspmatrix_csc(A):
        warnings.warn('CSC matrix format is required. Converting to CSC matrix.', scipy.sparse.SparseEfficiencyWarning)
        A = scipy.sparse.csc_matrix(A)

    A.sort_indices()
    A.eliminate_zeros()
    return A


def permutation_matrix(A):
    squared(A)
    assert A.nnz == A.shape[0]


def min_dtype(A, min_dtype):
    min_dtype = np.dtype(min_dtype)
    if A.dtype < min_dtype:
        logger.debug('Converting matrix {!r} to type {}.'.format(A, dtype))
        A = A.astype(dtype)
    return A


def index_dtype(A, dtype):
    if not (scipy.sparse.isspmatrix_csc(A) or scipy.sparse.isspmatrix_csr(A)):
        raise NotImplementedError("Only CSR and CSC are supported yet.")
    A.indices = np.asarray(A.indices, dtype=dtype)
    A.indptr = np.asarray(A.indptr, dtype=dtype)
    return A


import scipy.optimize
import numpy as np

import util.math.optimize.universal
import util.math.finite_differences

import util.logging
logger = util.logging.logger




def minimize(f, x0, jac=None, bounds=None, ineq_constraints=None, ineq_constraints_jac=None, global_method='random_start', global_iterations=1, global_stepsize=0.5, global_stepsize_update_interval=10, local_max_iterations=10**6, local_max_fun_evals=10**6):
    ## prepare input
    if jac is None:
        jac = lambda x: util.math.finite_differences.calculate(f, x, bounds=bounds, accuracy_order=2)[0]

    x0 = np.asarray(x0)
    if x0.ndim == 1:
        n = x0.shape[0]
        x0 = x0.reshape(1, n)
        m = 1
    else:
        n = x0.shape[1]
        m = x0.shape[0]

    if bounds is not None and len(bounds) != n:
        raise ValueError('x0 and bounds must have the same length, but their length are {} and {}.'.format(n, len(bounds)))

    if not global_method in ['random_start', 'basin_hopping']:
        raise ValueError('global_method {} has to be in {}'.format(global_method, 'random_start', 'basin_hopping'))

    ## chose local methods
    if ineq_constraints is not None:
        local_methods = ('SLSQP',)
    elif bounds is not None:
#         local_methods = ('SLSQP', 'L-BFGS-B', 'TNC')
        local_methods = ('L-BFGS-B', 'TNC')
    else:
        local_methods = ('SLSQP', 'BFGS', 'TNC')

    logger.debug('Glocal method {} chosen with local methods {}.'.format(global_method, local_methods))
    if bounds is not None:
        logger.debug('Using bounds {} for the optimization.'.format(bounds))

    ## prepare local options
    if ineq_constraints is None:
        constraints = ()
    else:
        constraints_dict = ({'type': 'ineq', 'fun': ineq_constraints})
        if ineq_constraints_jac is not None:
            constraints_dict['jac'] = ineq_constraints_jac
        constraints = [constraints_dict]

# #     #
#     if bounds is not None:
#         constraints = []
#         def lower_bound_contraint_dict(i):
#             def fun(x):
#                 return x[i] - bounds[i][0]
#             def jac(x):
#                 df = np.zeros(x.shape)
#                 df[i] = x[i]
#                 return df
#             return {'type': 'ineq', 'fun': fun, 'jac': jac}
#         def upper_bound_contraint_dict(i):
#             def fun(x):
#                 return bounds[i][1] - x[i]
#             def jac(x):
#                 df = np.zeros(x.shape)
#                 df[i] = - x[i]
#                 return df
#             return {'type': 'ineq', 'fun': fun, 'jac': jac}
#         for i in range(n):
#             constraints.append(lower_bound_contraint_dict(i))
#             constraints.append(upper_bound_contraint_dict(i))
# #     #

    disp = util.logging.is_debug()
    if disp:
        logger.debug('Debug output in optimization enabled.')
    else:
        logger.debug('Debug output in optimization disabled.')

    local_minimizer_options = {'method':'', 'jac':jac, 'bounds':bounds, 'constraints':constraints, 'options':{'maxiter': local_max_iterations, 'maxfun': local_max_fun_evals, 'disp':disp}}
#     local_minimizer_options = {'method':'', 'jac':jac, 'constraints':constraints, 'options':{'maxiter': local_max_iterations, 'maxfun': local_max_fun_evals, 'disp':disp}}
#     local_minimizer_options = {'method':'', 'jac':jac, 'bounds':bounds, 'options':{'maxiter': local_max_iterations, 'maxfun': local_max_fun_evals, 'disp':disp}}

    ## run optimization
    x_min = None
    f_min = np.inf

    ## random start point method
    if global_method == 'random_start':

        ## prepare start points
        if m < global_iterations and (bounds is None or not np.all(np.isfinite(bounds))):
            raise ValueError('Bounds must be finite if random start point method is chosen and the global iterations is greater then the number of starting points, but they are {}.'.format(bounds))

        x0s = np.empty([global_iterations, n])
        for i in range(global_iterations):
            if i < m:
                x0s[i] = x0[i]
            else:
                accept_x0 = False
                while not accept_x0:
                    for j in range(n):
                        x0s[i, j] = np.random.uniform(low=bounds[j][0], high=bounds[j][1], size=1)
                    accept_x0 = ineq_constraints(x0s[i]) >= 0

        ## run all local methods and from all start points
        for local_method in local_methods:
            local_minimizer_options['method'] = local_method
            for x0 in x0s:
                result = scipy.optimize.minimize(f, x0, **local_minimizer_options)
                logger.debug('Minimization with local method {} started from {} stopped with the following results: {}'.format(local_method, x0, result))

                if result.success and result.fun < f_min:
                    x_min = result.x
                    f_min = result.fun

    ## basin hopping method
    elif global_method == 'basin_hopping':
        x0 = x0[0]

        ## incorporate bounds and constarints
        if bounds is not None:
            global_bounds = np.empty([2, n])
            for i in range(n):
                if np.isreal(bounds[i][0]):
                    global_bounds[0, i] = bounds[i][0]
                else:
                    global_bounds[0, i] = - np.inf
                if np.isreal(bounds[i][1]):
                    global_bounds[1, i] = bounds[i][1]
                else:
                    global_bounds[1, i] = np.inf

            def bounds_test(**kwargs):
                x = kwargs['x_new']
                okay = bool(np.all(global_bounds[0] <= x) and np.all(x <= global_bounds[1]))
                return okay
        else:
            def bounds_test(**kwargs):
                return True

        if ineq_constraints is not None:
            def constraints_test(**kwargs):
                x = kwargs['x_new']
                okay = bounds_test(**kwargs) and bool(ineq_constraints(x) >= 0)
                return okay
        else:
            constraints_test = bounds_test

        ## run all local methods
        for local_method in local_methods:
            local_minimizer_options['method'] = local_method
            result = scipy.optimize.basinhopping(f, x0, minimizer_kwargs=local_minimizer_options, niter=global_iterations, disp=disp, accept_test=constraints_test, stepsize=global_stepsize, interval=global_stepsize_update_interval)
            logger.debug('Minimization basin hopping with local method {} started from {} stopped with the following results: {}'.format(local_method, x0, result))
            if result.fun < f_min:
                x_min = result.x
                f_min = result.fun


    ## return result
    if x_min is not None:
        return (x_min, f_min)
    else:
        raise util.math.optimize.universal.OptimizationError(result)




import warnings
import os.path

import numpy as np
import scipy.sparse
import scikits.sparse.cholmod

import util.math.matrix
import util.math.sparse.check

import util.logging
logger = util.logging.logger


RETURN_L = 'L'
RETURN_L_D = 'L_D'
RETURN_P_L = 'P_L'
RETURN_P_L_D = 'P_L_D'
CHOLMOD_ORDERING_METHODS = ('natural', 'default', 'best')



def approximate_positive_definite(A, min_abs_value=0, ordering_method='natural', min_diag_value=10**(-4), reorder_after_each_step=True, use_long=False, reduction_factors_file=None):
    logger.debug('Calculating positive definite approximation of matrix {!r} with min diagonal entry {}, min absolute value {}, ordering_method {}, reorder_after_each_step {}, use_long {} and reduction_factors_file {}.'.format(A, min_diag_value, min_abs_value, ordering_method, reorder_after_each_step, use_long, reduction_factors_file))
    
    assert min_diag_value > 0

    max_reduction_factor = 1 - 10**-6
    def multiply_off_diagonal_entries_with_factor(i, reduction_factor):
        # logger.debug('Multiplying off diagonal entries of row and column {} with {}.'.format(i, reduction_factor))
        
        ## check reduction_factor
        if not np.isfinite(reduction_factor):
            raise ValueError('Reduction factor {} with index {} has to be finite!'.format(reduction_factor, i))
        if not reduction_factor < 1:
            raise ValueError('Reduction factor {} with index {} has to be < 1!'.format(reduction_factor, i))
        if not reduction_factor >= 0:
            raise ValueError('Reduction factor {} with index {} has to be >= 0!'.format(reduction_factor, i))
        
        ## check max reduction factor
        if reduction_factor > max_reduction_factor:
            logger.warning('Current reduction factor {} for row and column {} is greater then max reduction factor {}. Past total reducing reduction factor is {}. Changing reduction factor to max reduction factor.'.format(reduction_factor, i, max_reduction_factor, reduction_factors[i]))
            reduction_factor = max_reduction_factor
        
        ## save reduction factor
        if reduction_factors_file is not None:
            reduction_factors[i] *= reduction_factor
            np.save(reduction_factors_file, reduction_factors)

        ## get indices
        A_ii = A[i,i]
        A_i_start_index = A.indptr[i]
        A_i_stop_index = A.indptr[i+1]
        assert A_i_stop_index - A_i_start_index > 1

        ## set column
        A.data[A_i_start_index:A_i_stop_index] *= reduction_factor

        ## set 0 for low values
        for k in range(A_i_start_index, A_i_stop_index):
            if np.abs(A.data[k]) < min_abs_value:
                A.data[k] = 0

        ## set row
        A_i_data = A.data[A_i_start_index:A_i_stop_index]
        A_i_rows = A.indices[A_i_start_index:A_i_stop_index]
        for j, A_ji in zip(A_i_rows, A_i_data):
            if i != j:
                A[i, j] = A_ji

        ## set diagonal entry
        A[i,i] = A_ii

        ## eliminate zeros
        if reorder_after_each_step or ordering_method == 'natural':
            A.eliminate_zeros()
    
    
    def get_p_i(i, f):        
        if ordering_method == 'natural':
            p_i = i
        else:
            p_i = f.P()[i]
        return p_i
        
    

    ## check input
    #TODO symmetry check
    A = util.math.sparse.check.sorted_squared_csc(A)
    A = util.math.sparse.check.min_dtype(A, np.float32)
    if use_long:
        A = util.math.sparse.check.index_dtype(A, np.int64)

    ## init
    n = A.shape[0]
    resolution = np.finfo(A.dtype).resolution
    if min_abs_value < resolution:
        logger.warning('Setting min abs value to resolution {} of data type.'.format(resolution))
        min_abs_value = resolution
    if min_diag_value < resolution:
        logger.warning('Setting min diag value to resolution {} of data type.'.format(resolution))
        min_diag_value = resolution

    ## apply reduction file values
    if reduction_factors_file is not None and os.path.exists(reduction_factors_file):
        reduction_factors = np.load(reduction_factors_file)
        for i in np.where((reduction_factors != 1))[0]:
            try:
                multiply_off_diagonal_entries_with_factor(i, reduction_factors[i])
            except ValueError as e:
                raise ValueError('Reduction factors file is wrong.') from e
    else:
        reduction_factors = np.ones(n, dtype=A.dtype)

    ## remove values below min abs
    mask = np.abs(A.data) < min_abs_value
    A.data[mask] = 0
    del mask
    A.eliminate_zeros()

    ## calculate positive definite approximation of A
    logger.debug('Checking if matrix is positive definite.')
    finished = False
    while not finished:
        
        ## compute cholesky factor
        try:
            try:
                f = scikits.sparse.cholmod.cholesky(A, ordering_method=ordering_method, use_long=use_long)
            except scikits.sparse.cholmod.CholmodTooLargeError as e:
                if not use_long:
                    warnings.warn('Problem to large for int, switching to long.')
                    return approximate_positive_definite(A, min_diag_value=min_diag_value, min_abs_value=min_abs_value, use_long=True)
                else:
                    raise
        except scikits.sparse.cholmod.CholmodNotPositiveDefiniteError as e:
            f = e.factor
        
        ## calculate LD and d
        LD = f.LD()     # Do not use f.L_D() -> higher memory consumption
        assert scipy.sparse.isspmatrix_csc(LD)
        d = LD.diagonal()
        
        ## check if all diagonal entries >= min_diag_value
        i_array = np.where(np.logical_not(np.logical_or(d >= min_diag_value, np.isclose(d, min_diag_value, atol=0, rtol=10**-4))))[0]
        finished = len(i_array) == 0
        
        ## if not, reduce off diagonal entries at row and column
        if not finished:
            ## get i and p_i
            i = i_array[0]
            p_i = get_p_i(i, f)
            del f, i_array
            
            ## check diagonal entry
            A_ii = A[p_i, p_i]
            if A_ii < min_diag_value:
                raise util.math.matrix.NoPositiveDefiniteMatrixError(A, 'Diagonal values of matrix must be at least {} to be able to ensure the minimum value, but the {}th entry is {}.'.format(min_diag_value, p_i, A_ii))
    
            ## get values for calulation of reduction factors
            assert scipy.sparse.isspmatrix_csc(LD)
            L_i = LD[i].tocsr()
            L_i_columns = L_i.indices
            L_i_data = L_i.data
            assert len(L_i_data) == len(L_i_columns) >= 2
            assert L_i_columns[-1] == i
    
            ## calculate reduction factor
            s = 0
            for j, L_ij in zip(L_i_columns[:-1], L_i_data[:-1]):
                assert j < i
                s += L_ij**2 * d[j]
            reduction_factor_i = ((A_ii - min_diag_value) / s)**(1/2)
        
            ## multiply off diagonal entries with reduction factor
            logger.debug('Row {} of cholesky decomposition, corresponding to row/column {} of input matrix, has diagonal value {} which is less than the min value {}. Multiplying off diagonal entries with {}.'.format(i, p_i, d[i], min_diag_value, reduction_factor_i))
            del LD, d, L_i, L_i_columns, L_i_data
            multiply_off_diagonal_entries_with_factor(p_i, reduction_factor_i)
            
    
    ## return
    A.eliminate_zeros()
    logger.debug('Returning reduction factors with average reduction factor {} and positive definite matrix {!r}.'.format(reduction_factors.mean(), A))
    return (A, reduction_factors)



def cholesky(A, ordering_method='default', return_type=RETURN_P_L, use_long=False):
    '''
    P A P' = L L'
    '''
    logger.debug('Calculating cholesky decomposition for matrix {!r} with ordering method {}, return type {} and use_long {}.'.format(A, ordering_method, return_type, use_long))

    ## check input
    return_types = (RETURN_L, RETURN_L_D, RETURN_P_L, RETURN_P_L_D)
    if ordering_method not in CHOLMOD_ORDERING_METHODS:
        raise ValueError('Unknown ordering method {}. Only values in {} are supported.'.format(ordering_method, CHOLMOD_ORDERING_METHODS))
    if return_type not in return_types:
        raise ValueError('Unknown return type {}. Only values in {} are supported.'.format(return_type, return_types))
        if ordering_method != 'natural' and return_type in (RETURN_L, RETURN_L_D):
            raise ValueError('Return type {} is only supported for "natural" ordering method.'.format(return_type))

    #TODO symmetry check
    A = util.math.sparse.check.sorted_squared_csc(A)

    ## calculate cholesky decomposition
    try:
        try:
            f = scikits.sparse.cholmod.cholesky(A, ordering_method=ordering_method, use_long=use_long)
        except scikits.sparse.cholmod.CholmodTooLargeError as e:
            if not use_long:
                warnings.warn('Problem to large for int, switching to long.')
                return cholesky(A, ordering_method=ordering_method, return_type=return_type, use_long=True)
            else:
                raise
    except scikits.sparse.cholmod.CholmodNotPositiveDefiniteError as e:
        raise util.math.matrix.NoPositiveDefiniteMatrixError(A, 'Row/column {} makes matrix not positive definite.'.format(e.column))
    del A

    ## calculate permutation matrix
    p = f.P()
    n = len(p)
    if return_type in (RETURN_P_L, RETURN_P_L_D):
        P = scipy.sparse.dok_matrix((n,n), dtype=np.int8)
        for i in range(n):
            P[i,p[i]] = 1
        P = P.tocsr()
        P.astype(np.int8)

    ## return P, L
    if return_type in (RETURN_L, RETURN_P_L):
        L = f.L().tocsr()
        if return_type == RETURN_L:
            assert np.all(p == np.arange(n))
            logger.debug('Returning lower triangular matrix {!r}.'.format(L))
            return (L,)
        else:
            logger.debug('Returning permutation matrix {!r} and lower triangular matrix {!r}.'.format(P, L))
            return (P, L)

    ## return P, L, D
    if return_type in (RETURN_L_D, RETURN_P_L_D):
        L, D = f.L_D()
        # Do not use f.L_D() -> higher memory consumption
        # LD = f.LD()
        if return_type == RETURN_L_D:
            logger.debug('Returning lower triangular matrix {!r} and diagonal matrix {!r}.'.format(P, L, D))
            return (L, D)
        else:
            logger.debug('Returning permutation matrix {!r}, lower triangular matrix {!r} and diagonal matrix {!r}.'.format(P, L, D))
            return (P, L, D)






import errno
import fnmatch
import os
import re
import stat
import time

import util.logging
logger = util.logging.logger


## file extension

def has_file_ext(file, ext):
    if not ext.startswith('.'):
        ext = '.' + ext
    return file.endswith(ext)


def add_file_ext_if_needed(file, ext):
    if has_file_ext(file, ext):
        return file
    else:
        if not ext.startswith('.'):
            ext = '.' + ext
        return file + ext


## get files

def find_with_condition_function(path, condition_function, exclude_files=False, exclude_dirs=False, use_absolute_filenames=False, recursive=False):
    ## use absolute path
    path = os.path.abspath(path)
    
    ## definde append filtered result function
    filtered_results = []
    
    def append_filtered(filename):
        if use_absolute_filenames:
            filename = os.path.join(path, filename)
        if condition_function(filename):
            filtered_results.append(filename)
    
    ## filter recursive
    if os.path.exists(path):
        if recursive:
            if not exclude_files:
                file_filter_function = append_filtered
            else:
                file_filter_function = None
            if not exclude_dirs:
                dir_filter_function = append_filtered
            else:
                dir_filter_function = None
            
            walk_all_in_dir(path, file_function=file_filter_function, dir_function=dir_filter_function, topdown=True, exclude_dir=True)
    
    ## filter not recursive
        else:
            for filename in os.listdir(path):
                append_filtered(filename)
            
            if exclude_files:
                if use_absolute_filenames:
                    filtered_results = [file for file in filtered_results if not os.path.isfile(file)]
                else:
                    filtered_results = [file for file in filtered_results if not os.path.isfile(os.path.join(path, file))]            
            if exclude_dirs:
                if use_absolute_filenames:
                    filtered_results = [file for file in filtered_results if not os.path.isdir(file)]
                else:
                    filtered_results = [file for file in filtered_results if not os.path.isdir(os.path.join(path, file))]     
    
    ## path is not existing
    else:
        filtered_results = []
    
    ## return 
    return filtered_results


def find_with_filename_pattern(path, filename_pattern, exclude_files=False, exclude_dirs=False, use_absolute_filenames=False, recursive=False):
    condition_function = lambda filename: fnmatch.fnmatch(filename, filename_pattern)
    filtered_results = find_with_condition_function(path, condition_function, exclude_files=exclude_files, exclude_dirs=exclude_dirs, use_absolute_filenames=use_absolute_filenames, recursive=recursive)
    return filtered_results


def find_with_regular_expression(path, regular_expression, exclude_files=False, exclude_dirs=False, use_absolute_filenames=False, recursive=False):
    regular_expression_object = re.compile(regular_expression)
    condition_function = lambda filename: regular_expression_object.fullmatch(filename)
    filtered_results = find_with_condition_function(path, condition_function, exclude_files=exclude_files, exclude_dirs=exclude_dirs, use_absolute_filenames=use_absolute_filenames, recursive=recursive)
    return filtered_results


def get_dirs(path=os.getcwd(), filename_pattern=None, use_absolute_filenames=False, recursive=False):
    if filename_pattern is None:
        filename_pattern = '*'
    dirs = find_with_filename_pattern(path, filename_pattern, exclude_files=True, use_absolute_filenames=use_absolute_filenames, recursive=recursive)
    return dirs

def get_files(path=os.getcwd(), filename_pattern=None, use_absolute_filenames=False, recursive=False):
    if filename_pattern is None:
        filename_pattern = '*'
    files = find_with_filename_pattern(path, filename_pattern, exclude_dirs=True, use_absolute_filenames=use_absolute_filenames, recursive=recursive)
    return files


## permissions

def file_mode(file):
    return os.stat(file)[stat.ST_MODE]


def file_permission(file):
    permission_octal_string = oct(file_mode(file))[-3:]
    permission_int = int(permission_octal_string, 8)
    return permission_int


def remove_write_permission(file):
    permission_old = file_mode(file)
    permission_new = permission_old
    for write_permission in (stat.S_IWUSR, stat.S_IWGRP, stat.S_IWOTH):
        permission_new = permission_new & ~ write_permission
    os.chmod(file, permission_new)
    logger.debug('Removing write permission of file {}. Mode changed from {} to {}.'.format(file, oct(permission_old)[-3:],oct( permission_new)[-3:]))


def add_write_permission(file):
    permission_old = file_mode(file)
    permission_new = permission_old
    for read_permission, write_permission in ((stat.S_IRUSR, stat.S_IWUSR), (stat.S_IRGRP, stat.S_IWGRP), (stat.S_IROTH, stat.S_IWOTH)):
        if permission_new & read_permission:
            permission_new = permission_new | write_permission
    os.chmod(file, permission_new)
    logger.debug('Adding write permission to file {}. Mode changed from {} to {}.'.format(file, oct(permission_old)[-3:], oct(permission_new)[-3:]))


def add_group_permissions(file, read=True, write=True, execute=True):
    permission_old = file_mode(file)
    permission_new = permission_old
    if read:
        permission_new = permission_new | stat.S_IRGRP
    if write:
        permission_new = permission_new | stat.S_IWGRP
    if execute:
        permission_new = permission_new | stat.S_IXGRP
    os.chmod(file, permission_new)
    logger.debug('Adding group permission (read={read}, write={write}, execute={execute}) to {file}. Mode changed from {permission_old} to {permission_new}.'.format(file=file, permission_old=oct(permission_old)[-3:], permission_new=oct(permission_new)[-3:], read=read, write=write, execute=execute))
    

def make_read_only(*files, not_exist_ok=False):
    for file in files:
        if not_exist_ok:
            try:
                remove_write_permission(file)
            except FileNotFoundError:
                logger.debug('File {} not existing, but this is okay.'.format(file))
                pass
        else:
            remove_write_permission(file)


def make_read_only_recursively(path, exclude_dir=True):
    logger.debug('Making recursively all files in {} read-only.'.format(path))
    file_function = lambda file: make_read_only(file)
    dir_function = lambda file: make_read_only(file)
    walk_all_in_dir(path, file_function, dir_function, exclude_dir=exclude_dir, topdown=False)


def make_writable(file, not_exist_ok=False):
    if not_exist_ok:
        try:
            add_write_permission(file)
        except FileNotFoundError:
            logger.debug('File {} does not exist, but this is okay.'.format(file))
            pass
    else:
        add_write_permission(file)


def make_writable_recursively(path, exclude_dir=True):
    logger.debug('Making recursively all files in {} writeable.'.format(path))
    file_function = lambda file: make_writable(file)
    dir_function = lambda file: make_writable(file)
    walk_all_in_dir(path, file_function, dir_function, exclude_dir=exclude_dir, topdown=False)


## remove

def _remove_general(remove_function, file, force=False, not_exist_okay=False):
    try:
        remove_function(file)
    except FileNotFoundError:
        if not not_exist_okay:
            raise
    except PermissionError:
        if force:
            (dir, filename) = os.path.split(file)
            make_writable(dir)
            remove_function(file)
        else:
            raise
    

def remove_dir(dir, force=False, not_exist_okay=False):
    _remove_general(os.rmdir, dir, force=force, not_exist_okay=not_exist_okay)


def remove_file(file, force=False, not_exist_okay=False):
    _remove_general(os.remove, file, force=force, not_exist_okay=not_exist_okay)


def remove_universal(file, force=False, not_exist_okay=False):
    try:
        remove_file(file, force=force, not_exist_okay=not_exist_okay)
    except IsADirectoryError:
        remove_dir(file, force=force, not_exist_okay=not_exist_okay)


def remove_recursively(dir, force=False, not_exist_okay=False, exclude_dir=False):
    try:
        remove_file(dir, force=force, not_exist_okay=not_exist_okay)
    except IsADirectoryError:
        walk_all_in_dir(dir, lambda file: remove_file(file, force=force, not_exist_okay=not_exist_okay), lambda dir: remove_dir(dir, force=force, not_exist_okay=not_exist_okay), exclude_dir=exclude_dir, topdown=False)


## utility functions

def flush_and_close(file):
    file.flush()
    os.fsync(file.fileno())
    file.close()
    while not os.path.exists(file.name):
        logger.warning('File {} is not available after flush and fsync. Waiting.'.format(file.name))
        time.sleep(1)


def walk_all_in_dir(dir, file_function=None, dir_function=None, topdown=True, exclude_dir=True):
    for (dirpath, dirnames, filenames) in os.walk(dir, topdown=topdown):
            if file_function is not None:
                for filename in filenames:
                    current_file = os.path.join(dirpath, filename)
                    file_function(current_file)
            if dir_function is not None:
                for dirname in dirnames:
                    current_dir = os.path.join(dirpath, dirname)
                    dir_function(current_dir)
    if not exclude_dir and dir_function is not None:
        dir_function(dir)


## fd functions

def fd_is_file(fd, file, not_exist_okay=False):
    ## get file stats
    try:
        file_stat = os.stat(file)
    except FileNotFoundError as e:
        if not_exist_okay:
            logger.debug('File {} does not exist, but this is okay.'.format(file))
            return False
        else:
            raise
    except OSError as e:
        if e.errno == errno.ESTALE and not_exist_okay:
            logger.debug('File {} does not exist, but this is okay.'.format(file))
            return False
        else:
            raise

    ## get fd stats
    try:
        fd_stat = os.fstat(fd)
    except OSError as e:
        if e.errno == errno.ESTALE and not_exist_okay:
            logger.debug('File referenced by file desciptor {} was removed, but this is okay.'.format(fd))
            return False
        else:
            raise
    
    ## check if same device and inode
    return fd_stat.st_dev == file_stat.st_dev and fd_stat.st_ino == file_stat.st_ino
     



import scipy.io

def save(file, array, value_name, oned_as='column'):
    scipy.io.savemat(file, {value_name: array}, oned_as=oned_as)

def load(file, value_name):
    return scipy.io.loadmat(file, squeeze_me=True)[value_name]


import fcntl
import os
import time

import util.io.filelock.general

import util.logging
logger = util.logging.get_logger()


# class FileLockTimeoutError(TimeoutError):
#     
#     def __init__(self, lock_filename, timeout, pid):
#         self.lock_filename = lock_filename
#         self.timeout = timeout
#         self.pid = pid
#         error_message = 'The lock {} could not be aquired by process {} within timeout {}.'.format(lock_filename, pid, timeout)
#         super().__init__(error_message)


class FileLock(object):

    def __init__(self, filename, timeout=None, sleep_time=0.5):
        self.filename = filename
        self.lock_filename = '{}.lock'.format(self.filename)
        self.timeout = timeout
        self.sleep_time = sleep_time
        self.pid = os.getpid()
        self.fd = None
    
    
    def lock_pid(self):
        """
        Get the pid of the lock.
        """
        
        if os.path.exists(self.lock_filename):
            return int(open(self.lock_filename).read())
        else:
            return None


    def is_locked_by_me(self):
        """
        See if the lock exists and is belongs to this process.
        """
        
        ## get pid of lock
        lock_pid = self.lock_pid()

        ## not locked
        if lock_pid is None:
            logger.debug('Lock {} is not aquired.'.format(self.lock_filename))
            return False
        
        ## locked by me
        if self.pid == lock_pid:
            logger.debug('Lock {} is aquired by me (pid: {}).'.format(self.lock_filename, self.pid))
            return True

        ## locked by alive process
        try:
            os.kill(lock_pid, 0)
            logger.debug('Lock {} is aquired by another (alive) process (pid: {}).'.format(self.lock_filename, lock_pid))
        
        ## locked by dead process
        except OSError:
            logger.debug('Lock {} is aquired by a dead process (pid: {}).'.format(self.lock_filename, lock_pid))
            try:
                lock_fd = os.open(self.lock_filename, os.O_RDWR)
            except FileNotFoundError:
                logger.debug('The lock is now removed by another process.')
            else:
                try:
                    fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    logger.debug('The lock will be removed by another process.')
                else:
                    if self.lock_pid() == lock_pid:
                        os.remove(self.lock_filename)
                        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
                        logger.debug('The lock is removed by me.')
                    else:
                        logger.debug('The lock is substituted by another lock.')
                os.close(lock_fd)
        return False


    def acquire_try_once(self):
        """
        Try to aquire the lock once.
        """
        
        if self.is_locked_by_me():
            return True
        else:
            try:
                self.fd = os.open(self.lock_filename, os.O_CREAT | os.O_RDWR | os.O_EXCL)
            except FileExistsError:
                logger.debug('I tried to get the lock {}, but another process was faster.'.format(self.lock_filename))
                return False
            else:
                os.write(self.fd, str(self.pid).encode('utf-8'))
                logger.debug('I got the lock {}.'.format(self.lock_filename))
                return True


    def acquire(self):
        """
        Try to aquire the lock.
        """
        
        if self.timeout is not None:
            sleep_intervals = int(self.timeout / self.sleep_time)
        else:
            sleep_intervals = float('inf')
        
        while not self.acquire_try_once() and sleep_intervals > 0:
            time.sleep(self.sleep_time)
            sleep_intervals -= 1
        
        if not self.is_locked_by_me():
            raise util.io.filelock.general.FileLockTimeoutError(self.lock_filename, self.timeout)
    

    def release(self):
        """
        Release the lock.
        """
        
        if self.is_locked_by_me():
            os.remove(self.lock_filename)
            logger.debug('The lock {} is released by me (pid: {}).'.format(self.lock_filename, self.pid))
        if self.fd:
            os.close(self.fd)
            self.fd = None


    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()

    def __del__(self):
        self.release()



class TestFileLock(object):
    
    def __init__(self):
        self.filename = 'test.txt'
        self.lock_filename = '{}.lock'.format(self.filename)

    def test_remove_dead_lock(self):
        with open(self.lockfile, 'w+') as f:
            f.write('9999999')
        assert os.path.exists(self.lockfile)
        with FileLock(self.filename):
            assert True

    def test_resume_lock(self):
        with open(self.lockfile, 'w+') as f:
            f.write(str(os.getpid()))
        with FileLock(self.filename, timeout=1):
            assert True

    def test_double_lock(self):
        with FileLock(self.filename):
            with FileLock(self.filename):
                assert True

    def test_exception_after_timeout(self):
        with open(self.lockfile, 'w+') as f:
            f.write('9999999')
        try:
            with FileLock(self.filename, timeout=1):
                assert False
        except FileLockTimeoutError:
            assert True





import numpy as np

import os

import util.index_database.general
import util.io.filelock.np

import util.logging
logger = util.logging.logger



class Database(util.index_database.general.Database):
    
    def __init__(self, array_file, value_reliable_decimal_places=15, tolerance_options=None):
        ## call super constructor
        super().__init__(value_reliable_decimal_places=value_reliable_decimal_places, tolerance_options=tolerance_options)
        
        ## set array file
        self.array_file = array_file
    
    
    def __str__(self):
        return 'Index array database {}'.format(self.locked_file.file)
    
    
    ## setter and getter for file
    
    @property
    def array_file(self):
        return self.locked_file.file
    
    @array_file.setter
    def array_file(self, array_file): 
        ## create dir if not existing
        os.makedirs(os.path.dirname(array_file), exist_ok=True)

        ## set array file
        self.locked_file = util.io.filelock.np.LockedFile(array_file)
    
    
    ## access
    
    def get_value(self, index):
        ## load database array
        try:
            db = self.locked_file.load()
        except FileNotFoundError:
            raise util.index_database.general.DatabaseIndexError(index)
        ## get value at index
        try:
            value = db[index]
        except IndexError:
            raise util.index_database.general.DatabaseIndexError(index)
        ## check if valid value is at index 
        if np.all(np.isnan(value)):
            raise util.index_database.general.DatabaseIndexError(index)
        else:
            return value
    
    
    def set_value(self, index, value, overwrite=False):
        logger.debug('{}: Setting value at index {} to {} with overwrite {}.'.format(self, index, value, overwrite))
        value = np.asanyarray(value)
        
        with self.locked_file.lock_object(exclusive=True):
            try:
                db = self.locked_file.load()
            except FileNotFoundError:
                db = value[np.newaxis]
            else:
                if index < len(db):
                    has_value = self.has_value(index)
                    if overwrite or not has_value:
                        db[index] = value
                    else:
                        logger.debug('{}: Overwritting value at index {} is not allowed.'.format(self, index))
                        raise util.index_database.general.DatabaseOverwriteError(index)
                else:
                    db_extension_len = index - len(db) + 1
                    db_extension = np.empty([db_extension_len, db.shape[1]]) * np.nan
                    db = np.concatenate([db, db_extension])
                    db[index] = value
            
            self.locked_file.save(db)


    def add_value(self, value):
        with self.locked_file.lock_object(exclusive=True):
            return super().add_value(value)


    def used_indices(self):
        try:
            db = self.locked_file.load()
        except FileNotFoundError:
            return ()
        else:
            used_mask = np.all(np.logical_not(np.isnan(db)), axis=1)
            used_indices = np.where(used_mask)[0]
            
            logger.debug('{}: Got {} used indices.'.format(self, len(used_indices)))
            return used_indices.astype(np.int32)
    
    
    def remove_index(self, index):
        logger.debug('{}: Removing index {}.'.format(self, index))
        
        with self.locked_file.lock_object(exclusive=True):
            if not self.has_value(index):
                raise util.index_database.general.DatabaseIndexError(index)
            
            db = self.locked_file.load()
            db[index] = db[index] * np.nan
            
            while len(db) > 0 and np.all(np.isnan(db[-1])):
                db = db[:-1]

            self.locked_file.save(db)
    

    def closest_indices(self, value):
        with self.locked_file.lock_object(exclusive=False):
            return super().closest_indices(value)


    def index(self, value):
        with self.locked_file.lock_object(exclusive=False):
            return super().index(value)





import numpy as np

import util.io.filelock.unix


class LockedFile(util.io.filelock.unix.LockedFile):

    def _load(self, file):
        return np.load(file)

    def _save(self, file, value):
        np.save(file, value)
    



import numpy as np
import scipy.sparse.linalg

import util.logging
logger = util.logging.logger


class CGMethodError(Exception):

    def __init__(self, exit_code):
        self.exit_code = exit_code

    def __str__(self):
        return 'CG method error. Illegal input or breakdown. Exit code {}.'.format(self.exit_code)


# k = 1
# xk_prev = None

def cg(A, b, x0=None, tol=10**-5, maxiter=None):
    logger.debug('CG method starting with tolerance {} and max iterations {}.'.format(tol, maxiter))

    ## chose x0
    if x0 is None:
        x0 = b

    ## callback -> output
    k = 1
    A_b = A * b
    # xk_prev = np.copy(x0)
    def callback(xk):
        nonlocal k
        # nonlocal xk_prev
        # diff_abs = np.abs(xk - xk_prev)
        # diff_rel = diff_abs / np.abs(xk_prev)
        #
        # logger.debug('Iteration {}: abs: avg diff = {:e}, averaged rel diff = {:e}, 2-norm abs diff = {:e}, 2-norm rel diff = {:e}'.format(k, diff_abs.mean(), diff_rel.mean(), np.linalg.norm(diff_abs, ord=2), np.linalg.norm(diff_rel, ord=2)))
        # xk_prev = np.copy(xk)

        A_xk = A * xk
        f = np.sum((xk- b) * (A_xk - A_b))
        residuum = np.abs(A_xk - b)
        logger.debug('Iteration {}: f = {:e}, avg residuum = {:e}, max residuum = {:e}, 2-norm residuum = {:e}'.format(k, f, residuum.mean(), residuum.max(), np.linalg.norm(residuum, ord=2)))
        k += 1

    ## run
    (x, exit_code) = scipy.sparse.linalg.cg(A, b, x0=x0, tol=tol, maxiter=maxiter, callback=callback)

    ## check exit code
    if exit_code == 0:
        logger.debug('CG method converged after {} iterations.'.format(k))
        return x
    elif exit_code > 0:
        logger.warning('CG method exited. Tolerance {} not achieved after {} iterations. Exit code {}'.format(tol, k, exit_code))
        return x
    else:
        e = CGMethodError(exit_code)
        logger.error(str(e))
        raise e



QUEUES = ('f_ocean', 'f_ocean2', 'express', 'small', 'medium', 'long', 'para_low')
MAX_WALLTIME = {'express': 3, 'small': 24, 'medium': 240, 'long': 480, 'para_low': 1000}

MODEL_RENAMING = {'python':'Python-3.3.2', 'python3':'Python-3.3.2', 'hdf5':'hdf5-1.8.16', 'matlab':'matlab2015a', 'petsc':'', 'intel14': 'intel1402', 'intelmpi14': 'intelmpi1402', 'intel16': 'intel16.0.0', 'intelmpi16': 'intelmpi16.0.0'}

COMMANDS = {'mpirun': 'mpirun -n {cpus:d} -machinefile $PBS_NODEFILE -r rsh {command}',
            'time': 'TIME_FMT="\nStatistics for %C:\nElapsed time: %Es, Exit code: %x\nCPU: %Us user mode, %Ss kernel mode, %P workload\nMemory: %Mkb max, %W swap outs\nContex switches: %c involuntarily, %w voluntarily\nPage faults: %F major, %R minor\nFile system I/O: %I inputs, %O outputs"\ncommand time -f "$TIME_FMT" {command}',
            'sub': '/opt/pbs/default/bin/qsub',
            'stat': '/opt/pbs/default/bin/qstat',
            'nodes': '/usr/local/bin/qnodes',
            'python': '/home/sunip229/local/bin/python3'}

NODE_INFOS = {'f_ocean2': {'nodes': 12, 'speed': 2.6, 'cpus': 16, 'memory': 128},
              'westmere': {'nodes': 18, 'speed': 2.67, 'cpus': 12, 'memory': 48, 'max_walltime': MAX_WALLTIME['para_low']},
              'shanghai': {'nodes': 26, 'speed': 2.4, 'cpus': 8, 'memory': 32, 'leave_free': 26, 'max_walltime': MAX_WALLTIME['para_low']},
              'amd128': {'nodes': 1, 'speed': 2.4, 'cpus': 16, 'memory': 128, 'leave_free': 1, 'max_walltime': MAX_WALLTIME['para_low']},
              'amd256': {'nodes': 1, 'speed': 2.1, 'cpus': 48, 'memory': 256, 'leave_free': 1, 'max_walltime': MAX_WALLTIME['para_low']},
              'fobigmem': {'nodes': 2, 'speed': 2.6, 'cpus': 24, 'memory': 1024, 'leave_free': 2}}




import numpy as np
import scipy.io

def save(file, A, precision=None):
    if precision is not None:
        precision_min = int(np.ceil(np.log10(max(A.shape))))
        if precision < precision_min:
            warnings.warn('Precision must be at least {1} to represent all indices. Precision changed from {2} to {1}.'.format(precision_min, precision))
            precision = precision_min

    scipy.io.mmwrite(file, A, precision=precision)

def load(file):
    return scipy.io.mmread(file)


class Observer:
    
    def notify(observable_name):
        raise NotImplementedError()


class UnknownObserverError(LookupError):
    
    def __init__(self, observer, observable_name=None):
        self.observer = observer
        self.observable_name = observable_name
        if observable_name is None:
            observable_name = 'all'
        message = 'The oberserver {} for {} is unknown'.format(observer, observable_name)
        super().__init__(message)



class Observable:
    
    def __init__(self):
        self._observers = {}
        self._everything_observers = []
        
    
    ## observer methods
    
    def add_observer(self, observer, observable_name=None):
        ## get everything observer list if no observable name is specified
        if observable_name is None:
            observer_list = self._everything_observers
        
        ## get specific observer list if observable name is specified
        else:
            try:
                observer_list = self._observers[observable_name]
            except KeyError:
                observer_list = []
                self._observers[observable_name] = observer_list
        
        ## add observer
        observer_list.append(observer)
        
    
    def remove_observer(self, observer, observable_name=None):
        ## get everything observer list if no observable name is specified
        if observable_name is None:
            observer_list = self._everything_observers
        
        ## get specific observer list if observable name is specified
        else:
            try:
                observer_list = self._observers[observable_name]
            except KeyError:
                raise UnknownObserverError(observer, observable_name=observable_name)
        
        ## remove observer
        try:
            observer_list.remove(observer)
        except ValueError:
            raise UnknownObserverError(observer, observable_name=observable_name)
        
        ## remove also list if empty
        if observable_name is not None and len(observer_listobserver_list) == 0:
            del self._observers[observable_name]

    
    def get_observers(self, observable_name=None, include_everything_observers=True):
        ## get specifiy observer
        if observable_name is not None:
            try:
                specific_observer = self._observers[observable_name]
            except KeyError:
                specific_observer = []
        
        ## get all observer
        if include_everything_observers:
            all_observer = self._everything_observers.copy()
        else:
            all_observer = []
        all_observer.extend(specific_observer)
        
        return all_observer
    
    
    def _notify_observers(self, observable_name=None, include_everything_observers=True):
        for observer in self.get_observers(observable_name=observable_name, include_everything_observers=include_everything_observers):
            observer(observable_name)
    
    
    def copy_observers_to(self, observable):
        observable._everything_observers.extend(self._everything_observers)

        for observable_name, observer in self._observers.items():
            observable.add_observer(observer, observable_name=observable_name)
    
    
    ## get and set observable_names method
    
    def _set_value(self, observable_name, new_value):
        raise NotImplementedError()
    
    def _del_value(self, observable_name):
        raise NotImplementedError()
        
    def _has_value(self, observable_name):
        raise NotImplementedError()
    
    def _get_value(self, observable_name):
        raise NotImplementedError()
    
    
    
    def _set_observable(self, observable_name, new_value):
        
        ## check old value
        if self._has_value(observable_name):
            old_value = self.new_value(observable_name)
            
            ## set only if different value
            must_set = np.any(new_value != old_value)
            if must_set:
                
                ## if values are observable_names with observers call associated observers of sub observable_names
                if isinstance(old_value, Observable) and isinstance(new_value, Observable):
                    ## copy observer
                    old_value.copy_observers_to(new_value)
                    
                    ## notify old observer
                    old_value._notify_observers(observable_name=None, include_everything_observers=True)
                    for observable_name in old_value._observers.keys():
                        old_has_value = old_value._has_value(observable_name)
                        new_has_value = new_value._has_value(observable_name)
                        if old_has_value != new_has_value or (old_has_value and new_has_value and np.any(old_value._get_value(observable_name) != new_value._get_value(observable_name))):
                            old_value._notify_observers(observable_name=observable_name, include_everything_observers=False)
        else:
            must_set = True
        
            
        ## set new observable_name value and call observer
        if must_set:
            self._set_value(observable_name, new_value)
            self._notify_observers(observable_name)
    
    
    def _del_observable(self, observable_name):
        self._del_value(observable_name)
        self._notify_observers(observable_name)



def observable_mutable_mapping_class(MutableMappingClass):
    
    class ObservableMutableMapping(Observable, MutableMappingClass):
        
        def __init__(self, *args, **kwargs):
            Observable.__init__(self)
            MutableMappingClass.__init__(self, *args, **kwargs)
        
        
        def __setitem__(self, key, value):
            self._set_observable(key, value)
    
        def __delitem__(self, key):
            self._del_observable(key)
        
        
        def _set_value(self, key, value):
            super().__setitem__(key, value)
        
        def _del_value(self, key):
            super().__delitem__(key)
            
        def _has_value(self, key):
            super().__contains__(key)
        
        def _get_value(self, key):
            return super().__getitem__(key)
        
    
    return ObservableMutableMapping



OberservableDict = observable_mutable_mapping_class(Dict)




def observable_attributes(original_class):
    class NewClass(object):
        
        def __init__(self, *args, **kwargs):
            self.original_instance = original_class(*args, **kwargs)
        
        def __getattribute__(self, s):
            """
            this is called whenever any attribute of a NewClass object is accessed. This function first tries to 
            get the attribute off NewClass. If it fails then it tries to fetch the attribute from self.original_instance (an
            instance of the decorated class). If it manages to fetch the attribute from self.original_instance, and 
            the attribute is an instance method then `time_this` is applied.
            """
            try:    
                x = super(NewClass, self).__getattribute__(s)
            except AttributeError:      
                pass
            else:
                return x
            x = self.original_instance.__getattribute__(s)
            if type(x) == type(self.__init__): # it is an instance method
                return time_this(x)                 # this is equivalent of just decorating the method with time_this
            else:
                return x
    return NewClass


import collections.abc

import numpy as np
import scipy.interpolate

import util.io.object
import util.parallel.with_multiprocessing
import util.logging
logger = util.logging.logger



def get_float_index_for_equidistant_values(value, value_range, dim):
    ## check input
    if value < value_range[0] or value > value_range[1]:
        raise ValueError('Value {} has to be in range {}.'.format(value, value_range))

    ## compute index
    index = (value - value_range[0]) / (value_range[1] - value_range[0]) * (dim)
    if index == dim:
        index = index - 1

    return index


def get_nearest_value_in_array(array, value):
    logger.debug('Getting nearest value in array for {}.'.format(value))

    distance = np.sum((array - value)**2,axis=-1)
    index = np.argmin(distance)
    nearest_value = array[index]

    logger.debug('Nearest value in array for {} is {}.'.format(value, nearest_value))

    return nearest_value


## interpolation

def data_with_float_index(data, index):
    ## get data at float index (linear) interpolated
    if len(index) == 0:
        return data
    else:
        index = np.asanyarray(index)
        lower = np.floor(index[0])
        upper = np.ceil(index[0])
        fraction = index[0] % 1
        return (1 - fraction) * data_with_float_index(data[lower], index[1:]) + fraction * data_with_float_index(data[upper], index[1:])


def data_with_regular_grid(data, points, point_ranges):
    ## make arrays
    data = np.asanyarray(data)
    points = np.asanyarray(points)

    ## check input
    assert data.ndim >= points.shape[1]

    ## make result array
    result_shape = (points.shape[0],) + data.shape[points.shape[1]:]
    results = np.empty(result_shape, dtype=data.dtype)

    ## convert points to indices
    point_ranges = np.asanyarray(point_ranges)
    normalized_points = (points - point_ranges[:,0]) / (point_ranges[:,1] - point_ranges[:,0])
    indice_shape = np.array(data.shape[:points.shape[1]])
    indices = normalized_points * (indice_shape - 1)

    ## interpolate
    for i in range(len(indices)):
        results[i] = data_with_float_index(data, indices[i])

    return results



def wrap_around(values, index, value_range_len, amount=1, return_also_indices=False):
    logger.debug('Wrapping around index {} with an amount of {} and value range len {}.'.format(index, amount, value_range_len))

    if amount < 0 or amount > 1:
        raise ValueError('Amount has to be between 0 and 1, but its {}.'.format(amount))

    ## prepare values
    values_len = len(values)
    values_len_amount = int(np.round(values_len * amount))

    logger.debug('Adding two times {} data values.'.format(values_len_amount-1))

    if amount != 0 and amount != 1:
        value_indices = np.argsort(values[:,index])
    else:
        value_indices = np.arange(values_len)

    ## concatenate
    if amount != 0:
        value_indices = np.concatenate((value_indices[- values_len_amount:], value_indices, value_indices[:values_len_amount]), axis=0)

        values = values[value_indices]
        values[:values_len_amount, index] -= value_range_len
        values[- values_len_amount:, index] += value_range_len

    ## return
    assert values.ndim >= 1
    assert values.shape[0] == values_len + 2 * values_len_amount
    assert value_indices.ndim == 1
    assert len(value_indices) == len(values)

    if return_also_indices:
        return values, value_indices
    else:
        return values


##

def change_dim(data, dim_index, new_dim):
    old_dim = data.shape[dim_index]
    if not (old_dim % new_dim == 0 or new_dim % old_dim == 0):
        raise ValueError('The dim {} of the data at index {} has to be a factor of the new_dim {} or vice versa.'.format(old_dim, dim_index, new_dim))

    new_shape = data.shape[:dim_index] + (new_dim,) + data.shape[dim_index+1:]
    assert data.ndim == len(new_shape)

    new_data = np.empty(new_shape, dtype=data.dtype)
    if new_dim < old_dim:
        step = int(old_dim / new_dim)
        logger.debug('Averaging dim {} to new dim {} with step {}.'.format(old_dim, new_dim, step))

        index_list = [Ellipsis,] * data.ndim
        for i in range(new_dim):
            index_list[dim_index] = slice(i * step, (i+1) * step)
            old_index = tuple(index_list)
            index_list[dim_index] = i
            new_index = tuple(index_list)
            new_data[new_index] = data[old_index].mean(axis=dim_index)

    elif new_dim > old_dim:
        step = int(new_dim / old_dim)
        logger.debug('Linear interpolating dim {} to new dim {} with step {}.'.format(old_dim, new_dim, step))

        index_list = [Ellipsis,] * data.ndim
        for i in range(new_dim):
            old_i_left = np.floor(i / step)
            old_i_right = old_i_left + 1
            if old_i_right == old_dim:
                old_i_right = 0
            old_i_fract = i / step - old_i_left

            index_list[dim_index] = old_i_left
            old_index_left = tuple(index_list)
            index_list[dim_index] = old_i_right
            old_index_right = tuple(index_list)
            index_list[dim_index] = i
            new_index = tuple(index_list)

            new_data[new_index] = (1 - old_i_fract) * data[old_index_left] + old_i_fract * data[old_index_right]

    else:
        new_data = data

    return new_data



## Base Interpolator

class Interpolator_Base():

    def __init__(self, data_points, data_values, method, possible_methods, scaling_values=None, copy_arrays=True):
        logger.debug('Initiating base interpolator with {} data points with scaling values {}, method {} and possible methods {}.'.format(len(data_points), scaling_values, method, possible_methods))

        if scaling_values is not None:
            if isinstance(scaling_values, collections.abc.Iterable):
                scaling_values = tuple(v if v is not None else 1 for v in scaling_values)
                scaling_values = np.array(scaling_values)
            else:
                scaling_values = np.array([scaling_values])
            assert scaling_values.ndim == 1

        self._scaling_values = scaling_values
        self._copy_arrays = np.asanyarray(copy_arrays)
        self._data_points = self._prepare_data_points(data_points)
        self._possible_methods = possible_methods
        self.method = method
        self.data_values = data_values

        assert self._scaling_values is None or len(self._scaling_values) in (1, self._data_points.shape[1])


    @property
    def data_points(self):
        return self._data_points

    @property
    def data_values(self):
        return self._data_values

    @data_values.setter
    def data_values(self, data_values):
        self._set_data_values(data_values)

    def _set_data_values(self, data_values):
        self._data_values = self._prepare_data_values(data_values)
        logger.debug('{} data values set.'.format(len(self._data_values)))

    @property
    def possible_methods(self):
        return self._possible_methods

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        if method not in self.possible_methods:
            raise ValueError('Method has to be in {}, but its value is {}.'.format(self.possible_methods, method))
        self._method = method

    @property
    def scaling_values(self):
        return self._scaling_values

    def _copy(self, index):
        copy_arrays = self._copy_arrays
        if copy_arrays.ndim == 0:
            return copy_arrays[()]
        elif len(copy_arrays) == 1:
            return copy_arrays[0]
        else:
            return copy_arrays[index]


    ## prepare values

    def _prepare_data_values(self, data_values):
        logger.debug('Preparing data values.')

        data_values = np.ascontiguousarray(data_values, dtype=np.double)

        if len(self.data_points) != len(data_values):
            raise ValueError('The data points and values must have the same first dimension, but its {} and {}.'.format(self.data_points.shape[0], data_values.shape[0]))
        if data_values.ndim != 1:
            raise ValueError('The data values must have one dimension, but its dimension is {}.'.format(data_values.ndim))

        assert data_values.ndim == 1
        assert data_values.shape[0] == self.data_points.shape[0]

        return data_values


    def _modify_data_points(self, points):
        points = self._scale(points)
        return points


    def _prepare_data_points(self, data_points):
        if self._copy(0):
            logger.debug('Preparing data points with a copy.')
            data_points = np.array(data_points, copy=True)
        else:
            logger.debug('Preparing data points.')
            data_points = np.asanyarray(data_points)

        if data_points.ndim == 1:
            data_points = data_points[None,:]
        elif data_points.ndim >= 3:
            raise ValueError('The data points must have two dimension, but its dimension is {}.'.format(data_points.ndim))
        if data_points.shape[0] == 0:
            raise ValueError('No data points passed.')
        if data_points.shape[1] < 2:
            raise ValueError('The data points must be at least 2 dimensional.')
        if self.scaling_values is not None and not len(self.scaling_values) in (1, data_points.shape[1]):
            raise ValueError('The length of the scaling values ({}) must be the number of dimensions of the data points ({}).'.format(len(self.scaling_values), data_points.shape[1]))

        data_points = self._modify_data_points(data_points)

        assert data_points.ndim == 2

        return data_points


    def _modify_interpolation_points(self, points):
        points = self._scale(points)
        return points


    def _prepare_interpolation_points(self, interpolation_points):
        if self._copy(1):
            logger.debug('Preparing interpolation points with a copy.')
            interpolation_points = np.array(interpolation_points, copy=True)
        else:
            logger.debug('Preparing interpolation points.')
            interpolation_points = np.asanyarray(interpolation_points)

        if interpolation_points.ndim == 1:
            interpolation_points = interpolation_points[None,:]
        elif interpolation_points.ndim >= 3:
            raise ValueError('Interpolation points have to be a vector or matrix array, but its shape is {}.'.format(points.shape))
        if self.data_points.shape[1] != interpolation_points.shape[1]:
            raise ValueError('The data and interpolation points must have the same second dimension, but its {} and {}.'.format(self.data_points.shape[1], interpolation_points.shape[1]))

        interpolation_points = self._modify_interpolation_points(interpolation_points)

        assert interpolation_points.ndim == 2
        assert self.data_points.shape[1] == interpolation_points.shape[1]

        return interpolation_points


    ## scale points
    
    def _scale(self, points):
        scaling_values = self.scaling_values

        if scaling_values is not None and np.any(scaling_values != 1):
            logger.debug('Scaling {} points with values {}.'.format(len(points), scaling_values))
            assert len(scaling_values) in (1, points.shape[1])
            points = points * scaling_values
        else:
            logger.debug('Do not scaling {} points'.format(len(points)))

        return points


    ## interpolate
    
    def _calculate_interpolation(self, interpolation_points):
        raise NotImplementedError("Please implement this method.")


    def interpolate(self, interpolation_points):
        interpolation_points_dim = len(interpolation_points)

        if interpolation_points_dim > 0:
            ## check input
            interpolation_points = self._prepare_interpolation_points(interpolation_points)

            ## interpolate
            logger.debug('Interpolating values with method {} at {} points from {} data points.'.format(self.method, interpolation_points_dim, len(self.data_values)))

            interpolated_values = self._calculate_interpolation(interpolation_points)

            number_of_interpolated_values = np.logical_not(np.isnan(interpolated_values)).sum()
            logger.debug('Values interpolated for {} points.'.format(number_of_interpolated_values))
        else:
            interpolated_values = np.array(())

        assert interpolated_values.ndim == 1
        assert interpolated_values.shape[0] == interpolation_points.shape[0]

        return interpolated_values


    ## save and load
    
    def save(self, file):
        util.io.object.save(file, self)
        logger.debug('Interpolator saved to {}.'.format(file))

    @staticmethod
    def load(file):
        interpolator = util.io.object.load(file)
        logger.debug('Interpolator loaded from {}.'.format(file))
        return interpolator



## Changeable interpolator

class Interpolator_Values_Changeable(Interpolator_Base):

    def __init__(self, data_points, data_values, method, scaling_values=None, copy_arrays=True):
        logger.debug('Initiating values changable interpolator with {} data points, scaling values {} and method {}.'.format(len(data_points), scaling_values, method))

        self._interpolator = None
        super().__init__(data_points, data_values, method, possible_methods=('nearest', 'linear'), scaling_values=scaling_values, copy_arrays=copy_arrays)


    def _set_data_values(self, data_values):
        super()._set_data_values(data_values)

        if self._interpolator is not None:
            if self.method == 'nearest':
                self._interpolator.values = data_values
            else:
                self._interpolator.values = data_values[:,None]
            logger.debug('Data values in {} interpolator updated.'.format(self.method))
        else:
            logger.debug('Data values in {} interpolator must not be updated, since interpolated is not constructed.'.format(self.method))


    def _get_interpolator(self):
        interpolator = self._interpolator

        if interpolator is None:
            logger.debug('Constructing {} interpolator.'.format(self.method))
            if self.method == 'nearest':
                tree_options = {'compact_nodes': False, 'balanced_tree': False}
                interpolator = scipy.interpolate.ndgriddata.NearestNDInterpolator(self.data_points, self.data_values, tree_options=tree_options)
            else:
                interpolator = scipy.interpolate.interpnd.LinearNDInterpolator(self.data_points, self.data_values)
            self._interpolator = interpolator
        else:
            logger.debug('Returning cached {} interpolator.'.format(self.method))

        assert callable(interpolator)

        return interpolator


    def _calculate_interpolation(self, interpolation_points):
        interpolator = self._get_interpolator()
        interpolated_values = interpolator(interpolation_points)
        return interpolated_values



## Changable and partitionable interpolator

class Interpolator_Values_Changeable_Partitionable(Interpolator_Base):

    def __init__(self, data_points, data_values, method='linear', number_of_interpolators=1, single_overlapping_amount=0.5, scaling_values=None, copy_arrays=True, parallel=False):
        logger.debug('Initiating partitionable interpolator with {} data points, scaling values {} and {} interpolators with single_overlapping_amount of {}.'.format(len(data_points), scaling_values, number_of_interpolators, single_overlapping_amount))

        self.parallel = parallel

        ## sort data by first index
        indices_sorted_by_first_dim = np.argsort(data_points[:,0])
        self._indices_sorted_by_first_dim = indices_sorted_by_first_dim
        data_points_sorted = data_points[indices_sorted_by_first_dim]
        data_values_sorted = data_values[indices_sorted_by_first_dim]

        self._interpolators = [None,] * number_of_interpolators
        copy_arrays = np.asanyarray(copy_arrays)
        if copy_arrays.ndim == 0:
            copy_interpolation_points = copy_arrays[()]
        elif len(copy_arrays) == 1:
            copy_interpolation_points = copy_arrays[0]
        else:
            copy_interpolation_points = copy_arrays[1]
        super().__init__(data_points_sorted, data_values, method, possible_methods=('nearest', 'linear'), scaling_values=scaling_values, copy_arrays=(False, copy_interpolation_points))


        ## compute interpolation bound values

        def get_last_index_of_same_value(values, index, step=1):
            value_len= len(values)
            if index < 0:
                index = 0
            elif index >= value_len:
                index = value_len - 1
            last_index = index
            next_index = last_index + step

            while next_index < value_len and next_index >= 0 and values[next_index] == values[last_index]:
                last_index = next_index
                next_index = last_index + step

            return last_index

        def get_last_index_of_same_value_vectorize(values, indices, step=1):
            last_indices = np.empty_like(indices, dtype=np.int)
            for i in range(len(indices)):
                last_indices[i] = get_last_index_of_same_value(values, indices[i], step=step)
            return last_indices


        interpolation_bound_indices = np.floor(np.arange(0, number_of_interpolators + 1) * data_points_sorted.shape[0] / (number_of_interpolators))
        interpolation_bound_indices = interpolation_bound_indices.astype(np.int)
        interpolation_bound_indices = get_last_index_of_same_value_vectorize(data_points_sorted[:,0], interpolation_bound_indices, step=-1)
        interpolation_bound_indices[-1] = data_points_sorted.shape[0] - 1

        logger.debug('The interpolation bounds indices for the partitioned interpolator are {}.'.format(interpolation_bound_indices))


        interpolation_bound_values = data_points_sorted[:,0][interpolation_bound_indices]
        interpolation_bound_values = np.asarray(interpolation_bound_values, dtype=np.float)
        interpolation_bound_values[0] = -np.inf
        interpolation_bound_values[-1] = np.inf

        self.interpolation_bound_values = interpolation_bound_values

        logger.debug('The interpolation bounds for the partitioned interpolator are {}.'.format(interpolation_bound_values))


        ## compute value range indices
        def get_interpolator_data_values_ranges(data_values_sorted_first_dim, interpolation_bound_indices, single_overlapping_amount):

            overlapping_len = len(data_values_sorted_first_dim) * single_overlapping_amount
            overlapping_interpolation_bound_indices_lower = np.floor(interpolation_bound_indices - overlapping_len)
            overlapping_interpolation_bound_indices_lower = overlapping_interpolation_bound_indices_lower.astype(np.int)
            overlapping_interpolation_bound_indices_upper = np.ceil(interpolation_bound_indices + overlapping_len)
            overlapping_interpolation_bound_indices_upper = overlapping_interpolation_bound_indices_upper.astype(np.int)

            number_of_interpolators = len(interpolation_bound_indices) - 1
            interpolator_data_value_ranges = np.empty((number_of_interpolators, 2), dtype=np.int)
            
            interpolator_data_value_ranges[:,0] = get_last_index_of_same_value_vectorize(data_values_sorted_first_dim, overlapping_interpolation_bound_indices_lower[:-1], step=-1)
            interpolator_data_value_ranges[:,1] = get_last_index_of_same_value_vectorize(data_values_sorted_first_dim, overlapping_interpolation_bound_indices_upper[1:], step=1)

            return interpolator_data_value_ranges

        data_value_range_indices = get_interpolator_data_values_ranges(data_points_sorted[:,0], interpolation_bound_indices, single_overlapping_amount)
        self.data_value_range_indices = data_value_range_indices

        logger.debug('The data value range indices for the partitioned interpolator are {}.'.format(data_value_range_indices))


    def _set_data_values(self, data_values):
        super()._set_data_values(data_values[self._indices_sorted_by_first_dim])

        for interpolator_index in range(self.number_of_interpolators):
            interpolator = self._interpolators[interpolator_index]
            if interpolator is not None:
                (index_start, index_end) = self.data_value_range_indices[interpolator_index]
                interpolator.data_values = self.data_values[index_start:index_end]
                logger.debug('Data values in interpolator with index {} with data from index {} to {} updated.'.format(interpolator_index, index_start, index_end))
            else:
                logger.debug('Data values in interpolator with index {} must not be updated, since interpolated is not constructed.'.format(interpolator_index))


    @property
    def number_of_interpolators(self):
        return len(self._interpolators)


    def _get_interpolator(self, interpolator_index):
        interpolator = self._interpolators[interpolator_index]

        if interpolator is None:
            logger.debug('Constructing interpolator with index {}.'.format(interpolator_index))

            (index_start, index_end) = self.data_value_range_indices[interpolator_index]
            interpolator = Interpolator_Values_Changeable(self.data_points[index_start:index_end], self.data_values[index_start:index_end], method=self.method, copy_arrays=False)
            self._interpolators[interpolator_index] = interpolator
        else:
            logger.debug('Returning cached interpolator with index {}.'.format(interpolator_index))

        return interpolator


    def _calculate_interpolation(self, interpolation_points):
        ## sort points by first dim
        indices_sorted_by_first_dim = np.argsort(interpolation_points[:,0])
        interpolation_points = interpolation_points[np.argsort(interpolation_points[:,0])]

        ## assign interpolation points to interpolators
        interpolation_points_dim = len(interpolation_points)
        interpolation_bound_values = self.interpolation_bound_values

        interpolator_index = 0
        start_index = 0
        end_index = 0

        end_indices = [0]

        number_of_interpolators = self.number_of_interpolators
        assert number_of_interpolators == len(interpolation_bound_values) - 1

        for interpolator_index in range(number_of_interpolators):
            while end_index < interpolation_points_dim and interpolation_points[end_index, 0] < interpolation_bound_values[interpolator_index + 1]:
                end_index += 1

            end_indices.append(end_index)
            start_index = end_index

        logger.debug('Interpolation points ranges {} assigned to interpolators {} with {} interpolator in use.'.format(end_indices, len(end_indices), number_of_interpolators))

        ## interpolate
        interpolated_values_sorted = np.empty(interpolation_points_dim)

        if not self.parallel or number_of_interpolators <= 1:
            logger.debug('Starting serial interpolation.')
            for interpolator_index in range(number_of_interpolators):
                interpolated_values_sorted[end_indices[interpolator_index]:end_indices[interpolator_index+1]] = self._calculate_interpolation_for_index(interpolator_index, interpolation_points[end_indices[interpolator_index]:end_indices[interpolator_index+1]])
        else:
            logger.debug('Starting interpolation with {} processes.'.format(number_of_interpolators))
            result = util.parallel.with_multiprocessing.map_parallel(self._calculate_interpolation_for_index_zipped, [(interpolator_index, interpolation_points[end_indices[interpolator_index]:end_indices[interpolator_index+1]]) for interpolator_index in range(number_of_interpolators)], number_of_processes=number_of_interpolators, chunksize=1)
            for interpolator_index in range(number_of_interpolators):
                interpolated_values_sorted[end_indices[interpolator_index]:end_indices[interpolator_index+1]] = result[interpolator_index]

        ## revert sort
        indices_sorted_by_first_dim_rev = np.empty(interpolation_points_dim, dtype=int)
        indices_sorted_by_first_dim_rev[indices_sorted_by_first_dim] = np.arange(interpolation_points_dim)

        assert np.all(indices_sorted_by_first_dim[indices_sorted_by_first_dim_rev] == np.arange(interpolation_points_dim))

        interpolated_values = interpolated_values_sorted[indices_sorted_by_first_dim_rev]
        return interpolated_values


    def _calculate_interpolation_for_index(self, interpolator_index, interpolation_points):
        interpolator = self._get_interpolator(interpolator_index)
        return interpolator.interpolate(interpolation_points)

    def _calculate_interpolation_for_index_zipped(self, zipped):
        interpolator_index, interpolation_points = zipped
        return self._calculate_interpolation_for_index(interpolator_index, interpolation_points)



## Nearest and linear Interpolator


class Interpolator_Nearest(Interpolator_Values_Changeable):
    def __init__(self, data_points, data_values, scaling_values=None, copy_arrays=True):
        super().__init__(data_points, data_values, method='nearest', scaling_values=scaling_values, copy_arrays=copy_arrays)

class Interpolator_Linear(Interpolator_Values_Changeable):
    def __init__(self, data_points, data_values, scaling_values=None, copy_arrays=True):
        super().__init__(data_points, data_values, method='linear', scaling_values=scaling_values, copy_arrays=copy_arrays)

class Interpolator_Linear_Partitionable(Interpolator_Values_Changeable_Partitionable):
    def __init__(self, data_points, data_values, scaling_values=None, copy_arrays=True, number_of_interpolators=1, single_overlapping_amount=0, parallel=True):
        super().__init__(data_points, data_values, method='linear', scaling_values=scaling_values, copy_arrays=copy_arrays, number_of_interpolators=number_of_interpolators, single_overlapping_amount=single_overlapping_amount, parallel=parallel)



## Universal interpolator

class Interpolator(Interpolator_Base):

    def __init__(self, data_points, data_values, method='linear_then_nearest_new_data', number_of_linear_interpolators=1, single_overlapping_amount_linear_interpolators=0, scaling_values=None, copy_arrays=True, parallel=False):

        logger.debug('Initiating interpolator {} with {} data points, scaling values {}, number of linear interpolators {} and single overlapping amount of linear interpolators {}.'.format(method, len(data_points), scaling_values, number_of_linear_interpolators, single_overlapping_amount_linear_interpolators))

        self.interpolators = []

        super().__init__(data_points, data_values, method=method, possible_methods=('linear_then_nearest_new_data', 'linear_then_nearest_same_data'), scaling_values=scaling_values, copy_arrays=copy_arrays)

        if number_of_linear_interpolators > 1:
            interpolator = Interpolator_Linear_Partitionable(self.data_points, self.data_values, number_of_interpolators=number_of_linear_interpolators, single_overlapping_amount=single_overlapping_amount_linear_interpolators, copy_arrays=False, parallel=parallel)
            self.interpolators.append(interpolator)
        elif number_of_linear_interpolators == 1:
            interpolator = Interpolator_Linear(self.data_points, self.data_values, copy_arrays=False)
            self.interpolators.append(interpolator)
        if number_of_linear_interpolators == 0 or method == 'linear_then_nearest_same_data':
            interpolator = Interpolator_Nearest(self.data_points, self.data_values, copy_arrays=False)
            self.interpolators.append(interpolator)

        assert len(self.interpolators) >= 1


    def _set_data_values(self, data_values):
        super()._set_data_values(data_values)

        for interpolator in self.interpolators:
            interpolator.data_values = self.data_values


    def _calculate_interpolation(self, interpolation_points):

        interpolation_values = self.interpolators[0].interpolate(interpolation_points)
        no_value_mask = np.isnan(interpolation_values)

        if np.any(no_value_mask):
            if self.method == 'linear_then_nearest_new_data':
                new_data_points = np.concatenate([self.data_points, interpolation_points[~ no_value_mask]])
                new_data_values = np.concatenate([self.data_values, interpolation_values[~ no_value_mask]])
                interpolator_nearest = Interpolator_Values_Changeable(self.data_points, self.data_values, method='nearest', copy_arrays=False)
            elif self.method == 'linear_then_nearest_same_data':
                interpolator_nearest = self.interpolators[1]

            interpolation_values[no_value_mask] = interpolator_nearest.interpolate(interpolation_points[no_value_mask])

        assert np.all(np.logical_not(np.isnan(interpolation_values)))

        return interpolation_values



class Periodic_Interpolator(Interpolator):

    def __init__(self, data_points, data_values, point_range_size, wrap_around_amount=None, method='linear_then_nearest_new_data', number_of_linear_interpolators=1, single_overlapping_amount_linear_interpolators=0, scaling_values=None, copy_arrays=True, parallel=False):

        logger.debug('Initiating periodic interpolator with point_range_size {} and wrap around amount {}.'.format(point_range_size, wrap_around_amount))

        point_range_size = np.asanyarray(point_range_size)
        wrap_around_amount = np.asanyarray(wrap_around_amount)

        if not (data_points.shape[1] == len(point_range_size) == len(wrap_around_amount)):
            raise ValueError('Second dim of data_points ({}) and len of point_range_size ({}) and len of wrap_around_amount ({}) have to be equal.'.format(data_points.shape[1], len(point_range_size), len(wrap_around_amount)))
        assert np.all(np.max(data_points, axis=0) - np.min(data_points, axis=0) <= point_range_size)

        self._point_range_size = point_range_size
        self._wrap_around_amount = wrap_around_amount
        self._data_indices = np.arange(len(data_points))

        super().__init__(data_points, data_values, method=method, number_of_linear_interpolators=number_of_linear_interpolators, single_overlapping_amount_linear_interpolators=single_overlapping_amount_linear_interpolators, scaling_values=scaling_values, copy_arrays=copy_arrays, parallel=parallel)

        assert len(self._data_points) == len(self._data_values) == len(self._data_indices)


    def _prepare_data_values(self, data_values):
        logger.debug('Wrapping around data values.')

        data_values = data_values[self._data_indices]
        data_values = super()._prepare_data_values(data_values)

        return data_values


    def _modify_points(self, points, is_data_points):
        for i in range(points.shape[1]):
            if self._wrap_around_amount is not None and self._wrap_around_amount[i] > 0:
                ## modulo for periodicity
                logger.debug('Calculate modulo {} of index {} of points.'.format(self._point_range_size[i], i))
                points[:, i] = points[:, i] % self._point_range_size[i]

                ## if data points, wrap around
                if is_data_points:
                    points, indices = util.math.interpolate.wrap_around(points, i, self._point_range_size[i], amount=self._wrap_around_amount[i], return_also_indices=True)
                    self._data_indices = self._data_indices[indices]

        return points


    def _modify_interpolation_points(self, points):
        points = self._modify_points(points, False)
        points = super()._modify_interpolation_points(points)
        return points


    def _modify_data_points(self, points):
        points = self._modify_points(points, True)
        points = super()._modify_data_points(points)
        return points



## universial interpolation method

def interpolate(data_points, data_values, interpolation_points, number_of_linear_interpolators=1, single_overlapping_amount_linear_interpolators=0, scaling_values=None):
    interpolator = Interpolator(data_points, data_values, number_of_linear_interpolators=number_of_linear_interpolators, single_overlapping_amount_linear_interpolators=single_overlapping_amount_linear_interpolators, scaling_values=scaling_values)
    return interpolator.interpolate(interpolation_points)



import sys

def multiprocessing_module():
    try:
        # globals()['multiprocessing']
        sys.modules['multiprocessing']
        return True
    except KeyError:
        return False


def scoop_module():
    try:
        import scoop
        return scoop.IS_RUNNING
    except ImportError:
        return False



import datetime

def datetime_to_float(dt):
    year_start = datetime.datetime(dt.year, 1, 1)
    next_year_start = datetime.datetime(dt.year + 1, 1, 1)
    float_fraction = (dt - year_start).total_seconds() / ((next_year_start - year_start).total_seconds())

    return dt.year + float_fraction


import random
import numpy as np

import deap.creator
import deap.base
import deap.tools
import deap.algorithms

import util.logging
logger = util.logging.logger





def init_individual_uniformly(bounds):
    bounds = np.asanyarray(bounds)
    individual_len = len(bounds)
    individual = np.empty(individual_len)
    for i in range(individual_len):
        individual[i] = random.uniform(bounds[i,0], bounds[i,1])
    return individual


def minimize(f, bounds, number_of_initial_individuals=100, number_of_generations=50, probability_of_crossover=0.5, probability_of_mutating=0.2, probability_of_mutating_in_each_dimension=0.5, crowding_degree_of_crossover=0.1, crowding_degree_of_mutation=0.1, tournament_size=3):
    logger.debug('Minimize function with {} initial individuals and {} generations.'.format(number_of_initial_individuals, number_of_generations))

    ## configure individual
    deap.creator.create("FitnessMin", deap.base.Fitness, weights=(-1.0,))
    deap.creator.create("Individual", np.ndarray, fitness=deap.creator.FitnessMin)

    ## configure population
    toolbox = deap.base.Toolbox()
    toolbox.register("init_individual_uniformly", init_individual_uniformly, bounds)
    toolbox.register("individual", deap.tools.initIterate, deap.creator.Individual, toolbox.init_individual_uniformly)
    toolbox.register("population", deap.tools.initRepeat, list, toolbox.individual)

    ## configure ea methods
    toolbox.register("evaluate", lambda x: (f(x),))
    toolbox.register("mate", deap.tools.cxSimulatedBinaryBounded, eta=crowding_degree_of_crossover, low=bounds[:,0].tolist(), up=bounds[:,1].tolist())
    toolbox.register("mutate", deap.tools.mutPolynomialBounded, eta=crowding_degree_of_mutation, low=bounds[:,0].tolist(), up=bounds[:,1].tolist(), indpb=probability_of_mutating_in_each_dimension)
    toolbox.register("select", deap.tools.selTournament, tournsize=tournament_size)

    ## configure stats
    stats = deap.tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", np.mean)
    stats.register("std", np.std)
    stats.register("min", np.min)
    stats.register("max", np.max)

    ## run optimization
    pop = toolbox.population(n=number_of_initial_individuals)
    hof = deap.tools.HallOfFame(1, similar=np.array_equal)
    pop, log = deap.algorithms.eaSimple(pop, toolbox, cxpb=probability_of_crossover, mutpb=probability_of_mutating, ngen=number_of_generations, stats=stats, halloffame=hof, verbose=True)
#     return pop, log, hof
    return hof[0]




# deap.creator.create("FitnessMin", deap.base.Fitness, weights=(-1.0,))
# deap.creator.create("Individual", np.ndarray, fitness=deap.creator.FitnessMin)
# toolbox = deap.base.Toolbox()
#
#
#
# # Attribute generator
# # toolbox.register("init_individual_uniformly", lambda :init_individual_uniformly(bounds))
# toolbox.register("init_individual_uniformly", init_individual_uniformly, bounds)
# # Structure initializers
# # toolbox.register("individual", deap.tools.initRepeat, deap.creator.Individual, toolbox.init_individual_uniformly, 100)
# toolbox.register("individual", deap.tools.initIterate, deap.creator.Individual, toolbox.init_individual_uniformly)
# toolbox.register("population", deap.tools.initRepeat, list, toolbox.individual)
#
#
# toolbox.register("evaluate", evalOneMax)
# #toolbox.register("mate", tools.cxTwoPoint)
# # toolbox.register("mate", tools.cxTwoPointCopy)
# # toolbox.register("mate", deap.tools.cxSimulatedBinaryBounded(ind1, ind2, eta, low.tolist(), up.tolist()))
# toolbox.register("mate", deap.tools.cxSimulatedBinaryBounded, eta=0.1, low=bounds[:,0].tolist(), up=bounds[:,1].tolist())
# # toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
# toolbox.register("mutate", deap.tools.mutPolynomialBounded, eta=0.1, low=bounds[:,0].tolist(), up=bounds[:,1].tolist(), indpb=0.05)
# # toolbox.register("select", deap.tools.selBest, k=20)
# toolbox.register("select", deap.tools.selTournament, tournsize=3)
#
# def main():
#     random.seed(64)
#     pop = toolbox.population(n=300)
#     #hof = tools.HallOfFame(1)
#     hof = deap.tools.HallOfFame(1, similar=np.array_equal)
#     stats = deap.tools.Statistics(lambda ind: ind.fitness.values)
#     stats.register("avg", np.mean)
#     stats.register("std", np.std)
#     stats.register("min", np.min)
#     stats.register("max", np.max)
#     pop, log = deap.algorithms.eaSimple(pop, toolbox, cxpb=0.5, mutpb=0.2, ngen=40, stats=stats, halloffame=hof, verbose=True)
#     return pop, log, hof
#
# if __name__ == "__main__":
#     main()
#
# if __name__ == "__main__":
#     random.seed(64)
#     bounds = np.array([[0,1],[4,9]])
#     def evalOneMax(individual):
#         return (sum(individual),)
#     x = minimize(evalOneMax, bounds)
#     print(x)




class FileLockTimeoutError(TimeoutError):
    
    def __init__(self, lock_filename, timeout):
        self.lock_filename = lock_filename
        self.timeout = timeout
        error_message = 'The lock for {} could not be aquired within timeout {}.'.format(lock_filename, timeout)
        super().__init__(error_message)



class FileLockNotAvailableError(Exception):
    
    def __init__(self, lock_filename, exclusive):
        self.lock_filename = lock_filename
        self.exclusive = exclusive
        if exclusive:
            lock_str = 'exclusive lock'
        else:
            lock_str = 'shared lock'
        error_message = 'The {} for {} was not available.'.format(lock_str, lock_filename)
        super().__init__(error_message)




import numpy as np


class MatrixError(Exception):

    def __init__(self, matrix, message=None, additional_message=None):
        self.matrix = matrix
        if message is None:
            message = 'Error with matrix {!r}!'.format(self.matrix)
        if additional_message is not None:
            message = message + ' ' + additional_message
        self.message = message

    def __str__(self):
        return self.message


class SingularMatrixError(MatrixError):

    def __init__(self, matrix, additional_message=None):
        message = 'Matrix {!r} is singular.'.format(matrix)
        super().__init__(matrix, message, additional_message)


class NoTriangularMatrixError(MatrixError):

    def __init__(self, matrix, additional_message=None):
        message = 'Matrix {!r} is no triangular matrix.'.format(matrix)
        super().__init__(matrix, message, additional_message)


class NoLeftTriangularMatrixError(MatrixError):

    def __init__(self, matrix, additional_message=None):
        message = 'Matrix {!r} is no left triangular matrix.'.format(matrix)
        super().__init__(matrix, message, additional_message)


class NoRightTriangularMatrixError(MatrixError):

    def __init__(self, matrix, additional_message=None):
        message = 'Matrix {!r} is no right triangular matrix.'.format(matrix)
        super().__init__(matrix, message, additional_message)


class NoPositiveDefiniteMatrixError(MatrixError):

    def __init__(self, matrix, additional_message=None):
        message = 'Matrix {!r} is not positive definite.'.format(matrix)
        super().__init__(matrix, message, additional_message)





def convert_to_matrix(array, dtype=None):
    array = np.asanyarray(array)
    matrix = np.asmatrix(array, dtype=dtype)
    if array.ndim == 1:
        matrix = matrix.T
    return matrix


def convert_matrix_to_array(matrix, dtype=None):
    array = np.asarray(matrix, dtype=dtype)
    if array.size == 1:
        array = array[0, 0]
    elif array.shape[1] == 1:
        array = array[:, 0]
    return array






import h5py

def load(file, name):
    with h5py.File(file, mode='r') as file:
        return file[name].value

def save(file, name, value):
    with h5py.File(file, mode='r+') as file:
        file[name][()] = value


def load_all_datasets(file, groupname):
    results = []

    with h5py.File(file, mode='r') as file:
        group = file[groupname]
        for dataset in group.keys():
            results.append(group[dataset].value)

    return results



import numpy as np
import scipy.sparse

import util.math.matrix
import util.math.sparse.check

import util.logging
logger = util.logging.logger


RETURN_L = 'L'
RETURN_L_D = 'L_D'


def cholesky(A, approximate_if_not_positive_semidefinite=False, return_type=RETURN_L_D):
    logger.debug('Calculating cholesky decomposition for matrix {!r}.'.format(A))

    ## check input
    return_types = (RETURN_L, RETURN_L_D)
    if return_type not in return_types:
        raise ValueError('Unknown return type {}. Only values in {} are supported.'.format(return_type, return_types))

    #TODO check symmetry of A
    # A = util.math.sparse.check.sorted_squared_csr(A)
    A = util.math.sparse.check.squared(A)
    A = scipy.sparse.tril(A).tocsr()

    ## init
    n = A.shape[0]

    L = scipy.sparse.lil_matrix((n,n))
    d = np.zeros(n)

    ## fill matrices
    for i in range(n):
        logger.debug('Calculating row {} of decomposition.'.format(i))

        positive_semi_definite = False
        reduction_factor = None
        while not positive_semi_definite:
            ## get A_i
            A_i_column_start = A.indptr[i]
            A_i_column_stop = A.indptr[i+1]
            A_i_columns = A.indices[A_i_column_start:A_i_column_stop]
            if reduction_factor is not None:
                A.data[A_i_column_start:A_i_column_stop] *= reduction_factor
            A_i_data = A.data[A_i_column_start:A_i_column_stop]
            A_i_pos = 0

            ## get L_i
            L_i_columns = L.rows[i]
            L_i_data = L.data[i]


            ## calculate row i of L
            for j in range(i):

                d_j = d[j]
                if d_j > 0:

                    # set L_ij = A[i,j]
                    while A_i_pos < len(A_i_columns) and A_i_columns[A_i_pos] < j:
                        A_i_pos += 1

                    if A_i_pos < len(A_i_columns) and A_i_columns[A_i_pos] == j:
                        L_ij = A_i_data[A_i_pos]
                    else:
                        L_ij = 0

                    # set L_ij -= sum(L_i * L_j * d)
                    L_j_columns = L.rows[j]
                    L_j_data = L.data[j]

                    L_i_pos = 0
                    L_j_pos = 0

                    while L_i_pos < len(L_i_columns) and L_j_pos < len(L_j_columns):
                        if L_i_columns[L_i_pos] == L_j_columns[L_j_pos]:
                            L_ij -= L_i_data[L_i_pos] * L_j_data[L_j_pos] *  d[L_i_columns[L_i_pos]]
                            L_i_pos += 1
                            L_j_pos += 1
                        elif L_i_columns[L_i_pos] < L_j_columns[L_j_pos]:
                            L_i_pos += 1
                        else:
                            L_j_pos += 1

                    # set L_ij /= d[j]
                    L_ij /= d_j

                    # set L[i,j] = L_ij
                    if L_ij != 0:
                        L[i,j] = L_ij


            ## calculate diagonal entry i

            # set d_i = A[i,i]
            while A_i_pos < len(A_i_columns) and A_i_columns[A_i_pos] < i:
                A_i_pos += 1
            A_ii = A_i_data[A_i_pos]

            if A_i_pos < len(A_i_columns) and A_i_columns[A_i_pos] == i:
                d_i = A_ii
            else:
                d_i = 0

            # set d_i -= sum(L_i**2 * d)
            for k in range(len(L_i_columns)):
                column_k = L_i_columns[k]
                d_i -= L_i_data[k]**2 * d[column_k]

            # set d[i] = d_i
            if d_i > 0:
                d[i] = d_i
                positive_semi_definite = True
                reduction_factor = None
            elif d_i < 0:
                if approximate_if_not_positive_semidefinite:
                    if A_ii > 0:
                        reduction_factor = (A_ii / (A_ii - d_i))**(1/2)
                    else:
                        reduction_factor = 0
                    logger.warn('Column/row {} makes matrix not positive definite. Diagonal entry would be {}. Multiplying row/column with {}.'.format(i, d_i, reduction_factor))
                else:
                    raise util.math.matrix.NoPositiveDefiniteMatrixError(A, 'Column/row {} of matrix is not positive definite. Diagonal entry would be {}.'.format(i, d_i))


    ## set 1 diagonal of L
    L.setdiag(1)

    ## convert to suitable sparse matrix formats
    logger.debug('Converting matrices to suitable formats.')
    L = L.tocsr()
    D = scipy.sparse.dia_matrix((d[np.newaxis,:], [0]), shape=(n,n))

    # ## return A
    # if return_types == RETURN_A:
    #     A = L * D * L.transpose()
    #     logger.debug('Returning positive definite matrix {!r}.'.format(A))
    #     return A

    ## return L
    if return_type == RETURN_L:
        D.data = D.data**(1/2)
        L = L*D
        logger.debug('Returning lower triangulat matrix {!r}.'.format(L))
        return L

    ## return L, D
    if return_type == RETURN_L_D:
        logger.debug('Returning lower triangulat matrix {!r} and diagonal matrix {!r}.'.format(L, D))
        return (L, D)






import os
import sys
import stat
import numpy as np
import re

import logging
logger = logging.getLogger(__name__)





# def get_sequence_from_values_or_file(value):
#     # check if float
#     try:
#         value = (float(value),)
#     # check if file
#     except (TypeError, ValueError):
#         try:
#             os.path.isfile(value)
#             try:
#                 value = np.loadtxt(value)
#             except (OSError, IOError) as e:
#                 raise ValueError('The value {} seams to be a file, but it could not been read.'.format(value)) from e
#     # check if sequence
#         except TypeError:
#             try:
#                 len(value)
#             except TypeError:
#                 raise ValueError('Value has to be a float, a sequence or a file containing a sequence.')
#
#     return value




## std out and std err

class String_List_Stream:
    def __init__(self, string_list, callback_stream):
        self.string_list = string_list
        self.callback_stream = callback_stream

    def write(self, message):
        self.string_list.append(message)
        self.callback_stream.write(message)


class Copy_Std_Streams:
    def __init__(self, string_list_out, string_list_err):
        self.string_list_out = string_list_out
        self.string_list_err = string_list_err

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.string_list_stream_out = String_List_Stream(self.string_list_out, self.old_stdout)
        self.string_list_stream_err = String_List_Stream(self.string_list_err, self.old_stderr)
        sys.stdout, sys.stderr = self.string_list_stream_out, self.string_list_stream_err
        self.old_stdout.flush(); self.old_stderr.flush()

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr







import warnings
import uuid

import numpy as np

import scoop
import scoop.futures

import util.parallel.universal

import util.logging
logger = util.logging.logger



def set_shared_object(object, name=None):
    if name is None:
        name = str(uuid.uuid4())

    logger.debug('Setting shared scoop object with name "{}".'.format(name))
    d = {name:object}
    scoop.shared.setConst(**d)

    return name


def get_shared_object(name):
    logger.debug('Getting shared scoop object "{}".'.format(name))
    return scoop.shared.getConst(name)





def map_parallel(function, values):
    assert callable(function)

    logger.debug('Calculating in parallel with scoop.')
#     results = scoop.futures.map(function, values) # Do not use map_as_completed, as it does not garanties the order.
    with warnings.catch_warnings():
        warnings.filterwarnings("error", module='scoop', category=RuntimeWarning, message='SCOOP was not started properly.')
        results = scoop.futures.map(function, values) # Do not use map_as_completed, as it does not garanties the order.
        results = tuple(results)

    logger.debug('Parallel calculation with {} results completed.'.format(len(results)))

    return results



def map_parallel_with_args(function, values, *args):
    logger.debug('Mapping function with {} args of types {} to values with scoop.'.format(len(args), tuple(map(type, args))))

    values = util.parallel.universal.args_generator_with_indices(values, (function,) + args)
    results = map_parallel(util.parallel.universal.eval_function_with_index_and_args, values)

    return reults





# def create_array(shape, function, function_args=None, index_position=1):
#     logger.debug('Creating array with shape {} in parallel with scoop.'.format(shape))
#
#     def function_arg_generator(shape, function_args, index_position):
#         index_position -= 1
#         if function_args is not None:
#             list_function_args = list(function_args)
#         else:
#             list_function_args = []
#         list_function_args.insert(index_position, None)
#
#         indices = np.ndindex(*shape)
#         for index in indices:
#             list_function_args[index_position] = index
#             yield tuple(list_function_args)
#
#
#     ## execute in parallel
#     results = map_parallel(function, function_arg_generator(shape, function_args, index_position))
#
#     ## create array
#     array = np.array(list(results))
#     logger.debug('Got {} results in parallel.'.format(len(array)))
#     array = array.reshape(shape)
#     return array
#
#
# # def create_array(shape, function, function_args=None, function_args_first=True):
# #     logger.debug('Creating array with shape {} in parallel with scoop.'.format(shape))
# #
# #     ## create indices
# #     indices = np.ndindex(*shape)
# #     if function_args is None:
# #         function_args = ()
# #     if function_args_first:
# #         indices = [(function_args, i) for i in indices]
# #     else:
# #         indices = [(i, function_args) for i in indices]
# #
# #     ## execute in parallel
# #     results = map(function, indices)
# #
# #     ## create array
# #     array = np.array(list(results))
# #     logger.debug('Got {} results in parallel.'.format(len(array)))
# #     array = array.reshape(shape)
# #     return array



# def is_scoop_loaded():
#     ## check if root process is running with scoop
#     try:
#         globals()['scoop']
#         return True
#     ## check if worker process
#     except KeyError:
#         return globals()['__name__']=='SCOOP_WORKER'



import numpy as np

import logging
logger = logging.getLogger(__name__)


def to_cartesian(points, measure='degree', surface_radius=None):

    logger.debug('Converting {} coordinates to cartesian coordinates with surface radius {}.'.format(measure, surface_radius))

    ## check input
    if measure not in ('degree', 'radian'):
        raise ValueError('Measure has to be "degree" or "radian" but it is {}.'.format(measure))
    if len(points.shape) != 2:
        raise ValueError('Points must have 2 dimensions but its shape is {}.'.format(points.shape))
    if points.shape[1] < 3:
        raise ValueError('Points second dimensions must be at least 3 but it is {}.'.format(points.shape[1]))


    ## unpack points
    (n, m) = points.shape
    longitude = points[:, -3]
    latitude = points[:, -2]
    r = points[:, -1]

    ## check input
    assert np.all(r >= 0)
    if measure == 'degree':
        assert (np.all(longitude >= -180) and np.all(longitude <= 180)) or (np.all(longitude >= 0) and np.all(longitude <= 360))
        assert (np.all(latitude >= -90) and np.all(latitude <= 90))
    else:
        assert (np.all(longitude >= -np.pi) and np.all(longitude <= np.pi)) or (np.all(longitude >= 0) and np.all(longitude <= 2*np.pi))
        assert (np.all(latitude >= -np.pi/2) and np.all(latitude <= np.pi/2))


    ## convert degree to radian
    if measure == 'degree':
        logger.debug('Converting degree to radian.')
        longitude = np.pi * longitude / 180
        latitude = np.pi * latitude / 180

    ## convert radius
    if surface_radius is not None:
        logger.debug('Setting radius to surface radius {} minus radius.'.format(surface_radius))
        r = surface_radius - r

    ## calculate cartesian
    logger.debug('Calulating cartesian coordinates.')

    long_sin = np.sin(longitude)
    long_cos = np.cos(longitude)
    lati_sin = np.sin(latitude)
    lati_cos = np.cos(latitude)

    x = r * lati_cos * long_cos
    y = r * lati_cos * long_sin
    z = r * lati_sin

    ## pack data
    x = x.reshape([n, 1])
    y = y.reshape([n, 1])
    z = z.reshape([n, 1])

    cartesian_points = np.concatenate((points[:,:-3], x, y, z),axis=1)

    return cartesian_points


import pickle
import pickletools

import util.io.fs



FILE_EXT = '.ppy'

def is_file(file):
    return util.io.fs.has_file_ext(file, FILE_EXT)

def add_file_ext(file):
    return util.io.fs.add_file_ext_if_needed(file, FILE_EXT)



def save(file, o, protocol=-1):
    file = add_file_ext(file)
    with open(file, 'wb') as file_object:
        pickle.dump(o, file_object, protocol=protocol)

def load(file):
    with open(file, 'rb') as file_object:
        o = pickle.load(file_object)
    return o


def protocol_version(file):
    maxproto = -1
    with open(file, 'rb') as file_object:
        for opcode, arg, pos in pickletools.genops(file_object):
            maxproto = max(maxproto, opcode.proto)
    return maxproto